(****
   contextual approximation for
   Realizability and Compositional Compiler Correctness 
      for a Polymorphic Language
   N Benton & C-K Hur
   April 2010
 ****)

Require Import Arith List Program FunctionalExtensionality ClassicalChoice Heq.
Require Import util. 
Require Import ssreflect ssrbool Heq Max.
Require Import cesd sysf sysfopsem logrel.

Set Implicit Arguments.

(* 
   Auxiliary Definitions
*)

Fixpoint VarTtoList u (E: Env u) A : (forall T, VarT E T -> A) -> list A :=
  match E return (forall T, VarT E T -> A) -> list A with
    | nil => fun rels => nil
    | T::E' => fun rels => 
      rels _ ZVAR :: VarTtoList (fun _ var => rels _ (SVAR _ var))
  end.

Fixpoint max_list (ns : list nat) : nat :=
  match ns with
    | nil => 0
    | n::ns' => max n (max_list ns')
  end.

Lemma VarTtoList_nth: forall u (E: Env u) T (var: VarT E T) A l (a:A),
  nth (idxVar var) (VarTtoList l) a = l T var.
Proof. by induction var; intros; simpl; [ | rewrite IHvar]. Qed.

Lemma le_max_list: forall ns k, nth k ns 0 <= max_list ns.
Proof.
induction ns; intros; first by destruct k.
simpl; destruct k; first by apply le_max_l.
by eapply le_trans; simpl; [apply IHns | apply le_max_r].
Qed.

Lemma f_app: forall A B (f g: A -> B) a, f = g -> f a = g a.
by intros; case H.
Defined.

Section TRelDefs_extension.

  Variable LeV : TRelV.
  Variable LeC : TRelC.

  Variable SLeV : STRelV.
  Variable SLeC : STRelC.

  Notation "tv <v< tv'" := (LeV tv tv') (at level 57).
  Notation "tv <c< tv'" := (LeC tv tv') (at level 57).
  Notation "tv =c= tv'" := (LeC tv tv' /\ LeC tv' tv) (at level 57).

  Notation "tv <<v<< tvs" := (SLeV tv tvs) (at level 57).
  Notation "tv <<c<< tvs" := (SLeC tv tvs) (at level 57).

  Definition TRel_app_beta := forall u E A B tv (tc : Exp u (A::E) B),
    APP (LAM tc) tv =c= SubExp {|tv|} tc.

  Lemma tybeta_C1: forall u (E:Env u) (T:Ty u) ty,
    (subEnv [|ty|] (subEnv shsub E), T) = (E, T).
  Proof. intros; simpl_sub. Qed.
  Implicit Arguments tybeta_C1 [[u] [E] [T] [ty]].

  Definition TRel_tyapp_beta := 
    forall u E T ty (tv : Val (S u) (subEnv shsub E) T),
    TAPP (TLAM tv) ty =c= ^<<SVal,PVal # tybeta_C1 >> subVal [|ty|] tv.

  Definition STRelV_ren := 
    forall u w (ren: renT u w) (E:Env u) T (tv: Val u E T) tvs,
    tv <<v<< tvs -> subVal (rts ren) tv <<v<< fun p => subVal (rts ren) (tvs p).
  Definition STRelC_ren :=
    forall u w (ren: renT u w) (E:Env u) T (tv: Exp u E T) tvs,
    tv <<c<< tvs -> subExp (rts ren) tv <<c<< fun p => subExp (rts ren) (tvs p).
  Definition STRelV_Ren :=
    forall u E E' (Ren: RenT E E') T (tv: Val u E T) tvs,
    tv <<v<< tvs -> 
    SubVal (RTS Ren) tv <<v<< fun p => SubVal (RTS Ren) (tvs p).
  Definition STRelC_Ren :=
    forall u E E' (Ren: RenT E E') T (tv: Exp u E T) tvs,
    tv <<c<< tvs -> 
    SubExp (RTS Ren) tv <<c<< fun p => SubExp (RTS Ren) (tvs p).

  Definition TRelV_trans := forall u E T (tv: Val u E T) tv' tv'',
    tv <v< tv' -> tv' <v< tv'' -> tv <v< tv''.
  Definition TRelC_trans := forall u E T (tv: Exp u E T) tv' tv'',
    tv <c< tv' -> tv' <c< tv'' -> tv <c< tv''.

  Definition TRelV_sub := 
    forall u w (sub: subT u w) (E:Env u) T (tv: Val u E T) tv',
    tv <v< tv' -> subVal sub tv <v< subVal sub tv'.
  Definition TRelC_sub :=
    forall u w (sub: subT u w) (E:Env u) T (tv: Exp u E T) tv',
    tv <c< tv' -> subExp sub tv <c< subExp sub tv'.
  Definition TRelV_Sub :=
    forall u E E' (Sub: SubT E E') Sub' T (tv: Val u E T) tv',
    (forall T (var: VarT E T), Sub _ var <v< Sub' _ var) ->
    tv <v< tv' -> 
    SubVal Sub tv <v< SubVal Sub' tv'.
  Definition TRelC_Sub :=
    forall u E E' (Sub: SubT E E') Sub' T (tv: Exp u E T) tv',
    (forall T (var: VarT E T), Sub _ var <v< Sub' _ var) -> 
    tv <c< tv' -> 
    SubExp Sub tv <c< SubExp Sub' tv'.
  
End TRelDefs_extension.


(*
   Definition of Context
*)

Inductive CtxV : forall u, Env u -> Ty u -> Type :=
| Ctx_PAIR1 : forall u (E:Env u) T1 T2, 
  CtxV E (T1***T2) -> Val _ E T2 -> CtxV E T1
| Ctx_PAIR2 : forall u (E:Env u) T1 T2, 
  CtxV E (T1***T2) -> Val _ E T1 -> CtxV E T2
| Ctx_INL : forall u (E:Env u) T1 T2, 
  CtxV E (T1+++T2) -> CtxV E T1
| Ctx_INR : forall u (E:Env u) T1 T2, 
  CtxV E (T1+++T2) -> CtxV E T2
| Ctx_TLAM : forall u (E:Env u) T,
  CtxV E (ALL T) -> CtxV (subEnv shsub E) T
| Ctx_PACK : forall u (E:Env u) T T',
  CtxV E (EXIST T) -> CtxV E (subTy [|T'|] T)
| Ctx_VAL: forall u (E:Env u) T,
  CtxC E T -> CtxV E T
| Ctx_OPN1 : forall u (E:Env u) (op: OperationNat),
  CtxC E INT -> Val _ E INT -> CtxV E INT
| Ctx_OPN2 : forall u (E:Env u) (op: OperationNat),
  CtxC E INT -> Val _ E INT -> CtxV E INT
| Ctx_OPB1 : forall u (E:Env u) (op: OperationBool),
  CtxC E BOOL -> Val _ E INT -> CtxV E INT
| Ctx_OPB2 : forall u (E:Env u) (op: OperationBool),
  CtxC E BOOL -> Val _ E INT -> CtxV E INT
| Ctx_FST : forall u (E:Env u) T1 T2,
  CtxC E T1 -> CtxV E (T1***T2)
| Ctx_SND : forall u (E:Env u) T1 T2,
  CtxC E T2 -> CtxV E (T1***T2)
| Ctx_CASE1 : forall u (E:Env u) T1 T2 T,
  CtxC E T -> Exp _ (T1::E) T -> Exp _ (T2::E) T -> CtxV E (T1+++T2)
| Ctx_APP1 : forall u (E:Env u) T1 T2,
  CtxC E T2 -> Val _ E T1 -> CtxV E (T1-->T2)
| Ctx_APP2 : forall u (E:Env u) T1 T2,
  CtxC E T2 -> Val _ E (T1-->T2) -> CtxV E T1
| Ctx_TAPP : forall u (E:Env u) T T',
  CtxC E (subTy [|T'|] T) -> CtxV E (ALL T)
| Ctx_UNPK1 : forall u (E:Env u) T1 T2,
  CtxC E T2 -> Exp _ (T1::subEnv shsub E) (subTy shsub T2) -> CtxV E (EXIST T1)
| CtxV_ren : forall u (E:Env u) T w (ren:renT u w),
  CtxV (subEnv (rts ren) E) (subTy (rts ren) T) -> CtxV E T
| CtxV_Ren : forall u (E:Env u) T E' (Ren:RenT E E'),
  CtxV E' T -> CtxV E T

with CtxC : forall u, Env u -> Ty u -> Type :=
| Ctx_ID : CtxC [] (@UNIT 0)
| Ctx_LAM : forall u (E:Env u) T1 T2, 
  CtxV E (T1-->T2) -> CtxC (T1::E) T2
| Ctx_REC : forall u (E:Env u) T1 T2, 
  CtxV E (T1-->T2) -> CtxC (T1::T1-->T2::E) T2
| Ctx_CASE2 : forall u (E:Env u) T1 T2 T,
  CtxC E T -> Val _ E (T1+++T2) -> Exp _ (T2::E) T -> CtxC (T1::E) T
| Ctx_CASE3 : forall u (E:Env u) T1 T2 T,
  CtxC E T -> Val _ E (T1+++T2) -> Exp _ (T1::E) T -> CtxC (T2::E) T
| Ctx_LET1 : forall u (E:Env u) T1 T2,
  CtxC E T2 -> Exp _ (T1::E) T2 -> CtxC E T1
| Ctx_LET2 : forall u (E:Env u) T1 T2,
  CtxC E T2 -> Exp _ E T1 -> CtxC (T1::E) T2
| Ctx_UNPK2 : forall u (E:Env u) T1 T2,
  CtxC E T2 -> Val _ E (EXIST T1) -> CtxC (T1::subEnv shsub E) (subTy shsub T2)
| CtxC_ren : forall u (E:Env u) T w (ren:renT u w),
  CtxC (subEnv (rts ren) E) (subTy (rts ren) T) -> CtxC E T
| CtxC_Ren : forall u (E:Env u) T E' (Ren:RenT E E'),
  CtxC E' T -> CtxC E T
.

Scheme CtxV_ind2 := Induction for CtxV Sort Prop
  with CtxC_ind2 := Induction for CtxC Sort Prop.

Combined Scheme CtxVC_ind from CtxV_ind2, CtxC_ind2.

Fixpoint pluginV u E T (ctx: CtxV E T) : Val u E T -> CExp UNIT :=
  match ctx in @CtxV u E T return Val u E T -> CExp UNIT with
  | Ctx_PAIR1 _ _ _ _ ctx tv2 => fun tvm => pluginV ctx (PAIR tvm tv2)
  | Ctx_PAIR2 _ _ _ _ ctx tv1 => fun tvm => pluginV ctx (PAIR tv1 tvm)
  | Ctx_INL _ _ _ _ ctx => fun tvm => pluginV ctx (INL _ tvm)
  | Ctx_INR _ _ _ _ ctx => fun tvm => pluginV ctx (INR _ tvm)
  | Ctx_TLAM _ _ _ ctx => fun tvm => pluginV ctx (TLAM tvm)
  | Ctx_PACK _ _ _ _ ctx => fun tvm => pluginV ctx (PACK tvm)
  | Ctx_VAL _ _ _ ctx => fun tvm => pluginC ctx (^tvm)
  | Ctx_OPN1 _ _ op ctx tv2 => fun tvm => pluginC ctx (OPN op tvm tv2)
  | Ctx_OPN2 _ _ op ctx tv1 => fun tvm => pluginC ctx (OPN op tv1 tvm)
  | Ctx_OPB1 _ _ op ctx tv2 => fun tvm => pluginC ctx (OPB op tvm tv2)
  | Ctx_OPB2 _ _ op ctx tv1 => fun tvm => pluginC ctx (OPB op tv1 tvm)
  | Ctx_FST _ _ _ _ ctx => fun tvm => pluginC ctx (FST tvm)
  | Ctx_SND _ _ _ _ ctx => fun tvm => pluginC ctx (SND tvm)
  | Ctx_CASE1 _ _ _ _ _ ctx tv1 tv2 => fun tvm => pluginC ctx (CASE tvm tv1 tv2)
  | Ctx_APP1 _ _ _ _ ctx tv2 => fun tvm => pluginC ctx (APP tvm tv2)
  | Ctx_APP2 _ _ _ _ ctx tv1 => fun tvm => pluginC ctx (APP tv1 tvm)
  | Ctx_TAPP _ _ _ T' ctx => fun tvm => pluginC ctx (TAPP tvm T')
  | Ctx_UNPK1 _ _ _ _ ctx tv2 => fun tvm => pluginC ctx (UNPK tvm tv2)
  | CtxV_ren _ _ _ _ ren ctx => fun tvm => pluginV ctx (subVal (rts ren) tvm)
  | CtxV_Ren _ _ _ _ Ren ctx => fun tvm => pluginV ctx (SubVal (RTS Ren) tvm)
  end

with pluginC u E T (ctx: CtxC E T) : Exp u E T -> CExp UNIT :=
  match ctx with
  | Ctx_ID => fun tvm => tvm
  | Ctx_LAM _ _ _ _ ctx => fun tvm => pluginV ctx (LAM tvm)
  | Ctx_REC _ _ _ _ ctx => fun tvm => pluginV ctx (REC tvm)
  | Ctx_CASE2 _ _ _ _ _ ctx tv tv2 => fun tvm => pluginC ctx (CASE tv tvm tv2)
  | Ctx_CASE3 _ _ _ _ _ ctx tv tv1 => fun tvm => pluginC ctx (CASE tv tv1 tvm)
  | Ctx_LET1 _ _ _ _ ctx tv2 => fun tvm => pluginC ctx (LET tvm tv2)
  | Ctx_LET2 _ _ _ _ ctx tv1 => fun tvm => pluginC ctx (LET tv1 tvm)
  | Ctx_UNPK2 _ _ _ _ ctx tv1 => fun tvm => pluginC ctx (UNPK tv1 tvm)
  | CtxC_ren _ _ _ _ ren ctx => fun tvm => pluginC ctx (subExp (rts ren) tvm)
  | CtxC_Ren _ _ _ _ Ren ctx => fun tvm => pluginC ctx (SubExp (RTS Ren) tvm)
  end.

(*
   Contextual Approximation
*)

(* CtxLeV and CtxLeC : 
   The usual notion of contextual equivalence*)
Definition CtxLeV : TRelV := fun _ _ _ tv tv' => 
  forall ctx, pluginV ctx tv |==> -> pluginV ctx tv' |==> .

Definition CtxLeC : TRelC := fun _ _ _ tv tv' =>
  forall ctx, pluginC ctx tv |==> -> pluginC ctx tv' |==> .

(* SCtxLeV and SCtxLeC :
   Contextual equivalence between a term and a sequence of terms *)
Definition ctx_adeqV u E T (tv:Val u E T) (tvs: stream (Val u E T)) :=
  forall ctx, pluginV ctx tv |==> -> exists p, pluginV ctx (tvs p) |==> .

Definition ctx_adeqC u E T (tv:Exp u E T) (tvs: stream (Exp u E T)) :=
  forall ctx, pluginC ctx tv |==> -> exists p, pluginC ctx (tvs p) |==> .

Definition SCtxLeV : STRelV := fun _ _ _ tv tvs =>
  is_chain (@CtxLeV _ _ _) tvs /\ ctx_adeqV tv tvs.

Definition SCtxLeC : STRelC := fun _ _ _ tv tvs =>
  is_chain (@CtxLeC _ _ _) tvs /\ ctx_adeqC tv tvs.

(*
   We assume well-known facts about the usual notion of contextual 
   equivalence, which are quite tedious to prove.
*)

Axiom CtxLe_beta: TRel_beta CtxLeC.
Axiom CtxLe_app_beta: TRel_app_beta CtxLeC.
Axiom CtxLe_tyapp_beta: TRel_tyapp_beta CtxLeC.
Axiom CtxLe_unwinding: 
  forall u (E: Env u) T1 T2 (tc: Exp u (T1::T1-->T2::E) T2) ctx,
  pluginV ctx (REC tc) |==> -> exists n, pluginV ctx (RECN n tc) |==> .
Axiom CtxLe_recn_monotone: 
  forall u (E: Env u) T1 T2 (tc: Exp u (T1::T1-->T2::E) T2) ctx i j,
    i <= j -> pluginV ctx (RECN i tc) |==> -> pluginV ctx (RECN j tc) |==> .

(*
   We also assume that if a value contextually approximates a chain of values
   as a compuation, then it contextually approximates the chain as a value.
   We omit the proof just because it is so tedious.
*)
Axiom SCtxLeC_value: forall u E T (tv: Val u E T) tvs,
  SCtxLeC (^tv) (fun p => ^tvs p) -> SCtxLeV tv tvs.

(*
   Definition of Adequate Compatiable Preorder 
*)

Record AdqCmpt (ParamV : TRelV) (ParamC : TRelC) := mkAdqCmpt
{
  AdqCmptV :> STRelV;
  AdqCmptC :> STRelC;

  (* Chain *)
  AdqCmptV_chain: STRelV_chain ParamV AdqCmptV;
  AdqCmptC_chain: STRelC_chain ParamC AdqCmptC;

  (* Reflexivity *)
  AdqCmptV_refl: STRelV_refl ParamV AdqCmptV;
  AdqCmptC_refl: STRelC_refl ParamC AdqCmptC;

  (* Renaming *)
  AdqCmptV_ren: STRelV_ren AdqCmptV;
  AdqCmptC_ren: STRelC_ren AdqCmptC;
  AdqCmptV_Ren: STRelV_Ren AdqCmptV;
  AdqCmptC_Ren: STRelC_Ren AdqCmptC;

  (* Compatibility *)
  AdqCmpt_VAR       : STRel_VAR AdqCmptV;
  AdqCmpt_UNIT      : STRel_UNIT AdqCmptV;
  AdqCmpt_INT       : STRel_INT AdqCmptV;
  AdqCmpt_PAIR      : STRel_PAIR AdqCmptV;
  AdqCmpt_INL       : STRel_INL AdqCmptV;
  AdqCmpt_INR       : STRel_INR AdqCmptV;
  AdqCmpt_LAM       : STRel_LAM AdqCmptV AdqCmptC;
  AdqCmpt_REC       : STRel_REC AdqCmptV AdqCmptC;
  AdqCmpt_TLAM      : STRel_TLAM AdqCmptV;
  AdqCmpt_PACK      : STRel_PACK AdqCmptV;
  AdqCmpt_VAL       : STRel_VAL AdqCmptV AdqCmptC;
  AdqCmpt_OPN       : STRel_OPN AdqCmptV AdqCmptC;
  AdqCmpt_OPB       : STRel_OPB AdqCmptV AdqCmptC;
  AdqCmpt_FST       : STRel_FST AdqCmptV AdqCmptC;
  AdqCmpt_SND       : STRel_SND AdqCmptV AdqCmptC;
  AdqCmpt_CASE      : STRel_CASE AdqCmptV AdqCmptC;
  AdqCmpt_LET       : STRel_LET AdqCmptC;
  AdqCmpt_APP       : STRel_APP AdqCmptV AdqCmptC;
  AdqCmpt_TAPP      : STRel_TAPP AdqCmptV AdqCmptC;
  AdqCmpt_UNPK      : STRel_UNPK AdqCmptV AdqCmptC;

  (* Adequacy *)
  AdqCmpt_adeq      : STRel_adeq AdqCmptC
}.

Hint Resolve AdqCmptV_refl AdqCmptC_refl.

Ltac eapply_AdqCmpt_chain := 
  try eapply AdqCmptV_chain; try eapply AdqCmptC_chain.

Ltac apply_AdqCmpt :=
  try apply AdqCmpt_VAR; try apply AdqCmpt_UNIT; 
  try apply AdqCmpt_INT; try apply AdqCmpt_PAIR; try apply AdqCmpt_INL;
  try apply AdqCmpt_INR; try apply AdqCmpt_LAM; try apply AdqCmpt_REC;
  try apply AdqCmpt_TLAM; try apply AdqCmpt_PACK; try apply AdqCmpt_VAL;
  try apply AdqCmpt_OPN; try apply AdqCmpt_OPB; try apply AdqCmpt_FST;
  try apply AdqCmpt_SND; try apply AdqCmpt_CASE; try apply AdqCmpt_LET;
  try apply AdqCmpt_APP; try apply AdqCmpt_TAPP; try apply AdqCmpt_UNPK;
  try apply AdqCmptV_ren; try apply AdqCmptC_ren;
  try apply AdqCmptV_Ren; try apply AdqCmptC_Ren;
  try apply AdqCmptV_refl; try apply AdqCmptC_refl;
  try apply AdqCmpt_adeq.

(****
   The union of all adequate congruences is an adequate congruence
 ****)

Definition SIdV : STRelV := fun _ _ _ tv tvs => tvs = cstr tv.
Definition SIdC : STRelC := fun _ _ _ tv tvs => tvs = cstr tv.
Definition IdV : TRelV := fun _ _ _ tv tv' => SIdV tv (cstr tv').
Definition IdC : TRelC := fun _ _ _ tv tv' => SIdC tv (cstr tv').

Lemma IdV_eq: forall u E T (tv tv': Val u E T), IdV tv tv' -> tv = tv'.
Proof. by intros; unfold IdV, SIdV, cstr in H; rewrite (f_app 0 H). Qed.

Lemma IdC_eq: forall u E T (tv tv': Exp u E T), IdC tv tv' -> tv = tv'.
Proof. by intros; unfold IdC, SIdC, cstr in H; rewrite (f_app 0 H). Qed.

Section Largest_Adq_Cmpt.

Variable   ParamLe : AdqCmpt IdV IdC.
Hypothesis ParamLeV_trans : STRelV_trans (AdqCmptV ParamLe).
Hypothesis ParamLeC_trans : STRelC_trans (AdqCmptC ParamLe).

Definition ParamLeV {u E T} (tv tv': Val u E T) : Prop := 
  AdqCmptV ParamLe tv (cstr tv').
Definition ParamLeC {u E T} (tv tv': Exp u E T) : Prop := 
  AdqCmptC ParamLe tv (cstr tv').

Definition AdqCmptP := AdqCmpt (@ParamLeV) (@ParamLeC).

Lemma AdqCmptV_cstr_refl: forall (rel: AdqCmptP) u E T (tv: Val u E T),
  AdqCmptV rel tv (cstr tv). 
Proof. 
intros; eapply (AdqCmptV_refl rel (tvs:=cstr tv) 0).
by red; intros; apply AdqCmptV_refl.
Qed.

Lemma AdqCmptC_cstr_refl: forall (rel: AdqCmptP) u E T (tv: Exp u E T),
  AdqCmptC rel tv (cstr tv). 
Proof. 
intros; eapply (AdqCmptC_refl rel (tvs:=cstr tv) 0).
by red; intros; apply AdqCmptC_refl.
Qed.

Lemma ParamLeV_refl: forall u E T (tv: Val u E T), ParamLeV tv tv.
Proof. 
by intros; change tv at 1 with ((cstr tv) 0); apply AdqCmptV_refl; repeat red.
Qed.

Lemma ParamLeC_refl: forall u E T (tv: Exp u E T), ParamLeC tv tv.
Proof. 
by intros; change tv at 1 with ((cstr tv) 0); apply AdqCmptC_refl; repeat red.
Qed.

Hint Resolve AdqCmptV_cstr_refl AdqCmptC_cstr_refl ParamLeV_refl ParamLeC_refl.

(*
   Identity Relation
*)
Definition IdAdqCmptV : STRelV := fun _ _ _ tv tvs => 
  is_chain ParamLeV tvs /\ exists p, ParamLeV tv (tvs p).

Definition IdAdqCmptC : STRelC := fun _ _ _ tv tvs =>
  is_chain ParamLeC tvs /\ exists p, ParamLeC tv (tvs p).

Definition IdAdqCmpt : AdqCmptP.
Proof.
apply (@mkAdqCmpt _ _ IdAdqCmptV IdAdqCmptC); try red; intros; try (by
  try destruct H as [? [? ?]]; try destruct H0 as [? [? ?]]; subst; auto;
  try (by split; [ | eexists 0]; red; intros; apply_AdqCmpt; Apply);
  try (by split; [ | eexists]; red; intros; apply_AdqCmpt; Apply);
  try (by split; [ | eexists (max x x0)]; red; intros; apply_AdqCmpt; 
       try (by Apply); 
       try (by first [eapply ParamLeV_trans|eapply ParamLeC_trans]; 
            intros; Apply; auto with arith))).
- destruct H as [? [p ?]]; destruct H0 as [? [p' ?]]; destruct H1 as [? [p'' ?]].
  split; [ | eexists (max p (max p' p''))]; red; intros; apply_AdqCmpt;
    try (by Apply);
    try (by eapply ParamLeV_trans; intros; Apply; auto with arith);
    by eapply ParamLeC_trans; intros; Apply; 
       eapply le_trans; last (by apply le_max_r); auto with arith.
- by destruct H as [? [p ?]]; eexists; apply (AdqCmpt_adeq ParamLe _ H1).
Defined.

(*
   Composition of Relations
*)

Definition composeSTRelV (rel: STRelV) (rels : stream STRelV) : STRelV :=
  fun _ _ _ tv tvs => exists tvs',
    rel _ _ _ tv tvs' /\ forall p, rels p _ _ _ (tvs' p) tvs.

Definition composeSTRelC (rel: STRelC) (rels : stream STRelC) : STRelC :=
  fun _ _ _ tv tvs => exists tvs',
    rel _ _ _ tv tvs' /\ forall p, rels p _ _ _ (tvs' p) tvs.

Definition composeAdqCmpt (rel: AdqCmptP) (rels: stream (AdqCmptP)) : AdqCmptP.
Proof.
intros; apply (@mkAdqCmpt _ _ (composeSTRelV rel rels) (composeSTRelC rel rels));
 try red; intros; 
 try destruct H as [tvs' [? ?]]; try destruct H0 as [tvs'' [? ?]];
 try destruct H1 as [tvs''' [? ?]];
 try (by eapply_AdqCmpt_chain; apply (H0 0));
 try (by eexists; split; [ | intros ]; apply_AdqCmpt).
- by eexists(fun p=>subVal (rts ren) (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>subExp (rts ren) (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>SubVal (RTS Ren) (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>SubExp (RTS Ren) (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>PAIR (tvs' p) (tvs'' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>INL _ (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>INR _ (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>LAM (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>REC (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>TLAM (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>PACK (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>^tvs' p); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>OPN op (tvs' p) (tvs'' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>OPB op (tvs' p) (tvs'' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>FST (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>SND (tvs' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>CASE (tvs' p) (tvs'' p) (tvs''' p));
     split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>LET (tvs' p) (tvs'' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>APP (tvs' p) (tvs'' p)); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>TAPP (tvs' p) _); split; [ |intros]; apply_AdqCmpt.
- by eexists(fun p=>UNPK (tvs' p) (tvs'' p)); split; [ |intros]; apply_AdqCmpt.
- by destruct (AdqCmpt_adeq rel _ H) as [p' ?]; 
     [do 2 eexists; Apply | apply (AdqCmpt_adeq _ _ (H1 p') H2)].
Defined.

(*
   Composition of List of Relations
*)

Fixpoint composeAdqCmptList (rels: list AdqCmptP) : AdqCmptP :=
  match rels with
    | nil => IdAdqCmpt
    | rel::rels' => composeAdqCmpt rel (cstr (composeAdqCmptList rels'))
  end.

Lemma composeAdqCmptList_extends: forall rels n,
  subrel_STRelV (AdqCmptV (nth n rels (IdAdqCmpt))) 
    (AdqCmptV (composeAdqCmptList rels)) /\
  subrel_STRelC (AdqCmptC (nth n rels (IdAdqCmpt)))
    (AdqCmptC (composeAdqCmptList rels)).
Proof.
induction rels; first by simpl; destruct n; split; red.
by move => n; split; (destruct n; simpl; red; intros;
[ exists tvs; split; [ |intros; apply_AdqCmpt; eapply_AdqCmpt_chain; Apply]
| exists (cstr tv); split; [apply_AdqCmpt | intros; eapply IHrels; Apply]]).
Qed.

(*
   The largest adequate congruence relation
*)

Definition UnionCmptV : STRelV := fun _ _ _ tv tvs =>
  exists rel : AdqCmptP, AdqCmptV rel tv tvs.

Definition UnionCmptC : STRelC := fun _ _ _ tv tvs =>
  exists rel : AdqCmptP, AdqCmptC rel tv tvs.

Lemma UnionCmptV_trans : STRelV_trans UnionCmptV.
Proof.
red; intros; destruct H as [rel ?].
replace_hyp H0 (choice _ H0).
destruct H0 as [rels ?].
by exists (composeAdqCmpt rel rels); exists tvs.
Qed.

Lemma UnionCmptC_trans : STRelC_trans UnionCmptC.
Proof.
red; intros; destruct H as [rel ?].
replace_hyp H0 (choice _ H0).
destruct H0 as [rels ?].
by exists (composeAdqCmpt rel rels); exists tvs.
Qed.

Definition UnionCmpt : AdqCmptP.
apply (@mkAdqCmpt _ _ UnionCmptV UnionCmptC); try red; intros;
  try destruct H as [rel ?]; try destruct H0 as [rel' ?];
  try (by eapply_AdqCmpt_chain; Apply);
  try (by exists IdAdqCmpt; apply_AdqCmpt);
  try (by exists rel; apply_AdqCmpt);
  try (by exists (composeAdqCmptList (rel::rel'::[])); apply_AdqCmpt;
          [ apply (composeAdqCmptList_extends _ 0) 
          | apply (composeAdqCmptList_extends _ 1) ];Apply).
- destruct H1 as [rel'' ?].
  by exists (composeAdqCmptList (rel::rel'::rel''::[])); apply_AdqCmpt;
  [ apply (composeAdqCmptList_extends _ 0) 
  | apply (composeAdqCmptList_extends _ 1) 
  | apply (composeAdqCmptList_extends _ 2) ];Apply.
- by eapply AdqCmpt_adeq; [ | eexists]; Apply.
Defined.

Lemma subrel_ParamLeV: forall u E T (tv tv': Val u E T),
  ParamLeV tv tv' -> to_TRelV (AdqCmptV UnionCmpt) tv tv'.
Proof. 
red; intros; exists IdAdqCmpt.
by split; [red; intros; apply_AdqCmpt | exists 0]. 
Qed.

Lemma subrel_ParamLeC: forall u E T (tv tv': Exp u E T),
  ParamLeC tv tv' -> to_TRelC (AdqCmptC UnionCmpt) tv tv'.
Proof. 
red; intros; exists IdAdqCmpt.
by split; [red; intros; apply_AdqCmpt | exists 0].
Qed.

End Largest_Adq_Cmpt.

(*****
   Compatibility Closure
 *****)

Section Cmpt_Closure.

Variable ParamLe : AdqCmpt IdV IdC.
Variable SLeV : STRelV.
Variable SLeC : STRelC.

Hypothesis ParamLeV_trans : STRelV_trans (AdqCmptV ParamLe).
Hypothesis ParamLeC_trans : STRelC_trans (AdqCmptC ParamLe).
Hypothesis ParamLeV_ctx_chain : forall u E T (tvs: stream (Val u E T)),
  is_chain (ParamLeV ParamLe) tvs -> is_chain (@CtxLeV _ _ _) tvs.
Hypothesis ParamLeC_ctx_chain : forall u E T (tvs: stream (Exp u E T)),
  is_chain (ParamLeC ParamLe) tvs -> is_chain (@CtxLeC _ _ _) tvs.

Hypothesis SLeV_chain : STRelV_chain (@ParamLeV ParamLe) SLeV.
Hypothesis SLeC_chain : STRelC_chain (@ParamLeC ParamLe) SLeC.
Hypothesis SLeV_ctx_adeq: forall u E T (tv: Val u E T) tvs,
  SLeV tv tvs -> ctx_adeqV tv tvs.
Hypothesis SLeC_ctx_adeq: forall u E T (tv: Exp u E T) tvs,
  SLeC tv tvs -> ctx_adeqC tv tvs.

Inductive CmptClosV : STRelV :=
| CClosV_refl  : STRelV_refl (@ParamLeV ParamLe) CmptClosV
| CClosV_ren   : STRelV_ren CmptClosV
| CClosV_Ren   : STRelV_Ren CmptClosV
| CClos_VAR    : STRel_VAR CmptClosV
| CClos_UNIT   : STRel_UNIT CmptClosV
| CClos_INT    : STRel_INT CmptClosV
| CClos_PAIR   : STRel_PAIR CmptClosV
| CClos_INL    : STRel_INL CmptClosV
| CClos_INR    : STRel_INR CmptClosV
| CClos_LAM    : STRel_LAM CmptClosV CmptClosC
| CClos_REC    : STRel_REC CmptClosV CmptClosC
| CClos_TLAM   : STRel_TLAM CmptClosV
| CClos_PACK   : STRel_PACK CmptClosV
| CClosV_base  : forall u E T (tv:Val u E T) tvs,
  SLeV tv tvs -> CmptClosV tv tvs

with CmptClosC : STRelC :=
| CClosC_refl  : STRelC_refl (@ParamLeC ParamLe) CmptClosC
| CClosC_ren   : STRelC_ren CmptClosC
| CClosC_Ren   : STRelC_Ren CmptClosC
| CClos_VAL    : STRel_VAL CmptClosV CmptClosC
| CClos_OPN    : STRel_OPN CmptClosV CmptClosC
| CClos_OPB    : STRel_OPB CmptClosV CmptClosC
| CClos_FST    : STRel_FST CmptClosV CmptClosC
| CClos_SND    : STRel_SND CmptClosV CmptClosC
| CClos_CASE   : STRel_CASE CmptClosV CmptClosC
| CClos_LET    : STRel_LET CmptClosC
| CClos_APP    : STRel_APP CmptClosV CmptClosC
| CClos_TAPP   : STRel_TAPP CmptClosV CmptClosC
| CClos_UNPK   : STRel_UNPK CmptClosV CmptClosC
| CClosC_base  : forall u E T (tv:Exp u E T) tvs, 
  SLeC tv tvs -> CmptClosC tv tvs
.

Scheme CmptClosV_ind2 := Induction for CmptClosV Sort Prop
  with CmptClosC_ind2 := Induction for CmptClosC Sort Prop.

Combined Scheme CmptClosVC_ind from CmptClosV_ind2, CmptClosC_ind2.

(* CmptV is chain *)

Lemma CClos_chain:
  STRelV_chain (@ParamLeV ParamLe) CmptClosV /\
  STRelC_chain (@ParamLeC ParamLe) CmptClosC.
Proof.
apply CmptClosVC_ind; intros; auto; 
  try (by red; intros; apply_AdqCmpt; intros; Apply);
  try (by first [eapply SLeV_chain | eapply SLeC_chain]; Apply).
Qed.

Definition CClosV_chain := proj1 CClos_chain.
Definition CClosC_chain := proj2 CClos_chain.

(* CmptV is adequate *)

Lemma CClos_ctx_adeq:
  (forall u E T (tv: Val u E T) tvs, CmptClosV tv tvs -> ctx_adeqV tv tvs) /\
  (forall u E T (tv: Exp u E T) tvs, CmptClosC tv tvs -> ctx_adeqC tv tvs).
Proof.
apply CmptClosVC_ind; try red; intros;
  try rename c into Hc; try rename c0 into Hc0; try rename c1 into Hc1; 
  try rename s into Hs; try (by eexists 0); try (by eexists p).
- by destruct (H (CtxV_ren _ _ ren ctx) H0) as [p ?]; exists p.
- by destruct (H (CtxV_Ren Ren ctx) H0) as [p ?]; exists p.
- destruct (H (Ctx_PAIR1 ctx tv2) H1) as [p ?].
  destruct (H0 (Ctx_PAIR2 ctx (tvs1 p)) H2) as [p' ?].
  by exists (max p p');
   apply (ParamLeV_ctx_chain (CClosV_chain Hc) (le_max_l _ _) (Ctx_PAIR1 _ _));
   apply (ParamLeV_ctx_chain (CClosV_chain Hc0) (le_max_r _ _) (Ctx_PAIR2 _ _)).
- by destruct (H (Ctx_INL ctx) H0) as [p ?]; exists p.
- by destruct (H (Ctx_INR ctx) H0) as [p ?]; exists p.
- by destruct (H (Ctx_LAM ctx) H0) as [p ?]; exists p.
- by destruct (H (Ctx_REC ctx) H0) as [p ?]; exists p.
- by destruct (H (Ctx_TLAM ctx) H0) as [p ?]; exists p.
- by destruct (H (Ctx_PACK _ ctx) H0) as [p ?]; exists p.
- by apply (SLeV_ctx_adeq Hs _ H).
- by destruct (H (CtxC_ren _ _ ren ctx) H0) as [p ?]; exists p.
- by destruct (H (CtxC_Ren Ren ctx) H0) as [p ?]; exists p.
- by destruct (H (Ctx_VAL ctx) H0) as [p ?]; exists p.
- destruct (H (Ctx_OPN1 _ ctx tv2) H1) as [p ?].
  destruct (H0 (Ctx_OPN2 _ ctx (tvs1 p)) H2) as [p' ?].
  by exists (max p p');
   apply (ParamLeV_ctx_chain (CClosV_chain Hc) (le_max_l _ _) (Ctx_OPN1 _ _ _));
   apply (ParamLeV_ctx_chain (CClosV_chain Hc0) (le_max_r _ _) (Ctx_OPN2 _ _ _)).
- destruct (H (Ctx_OPB1 _ ctx tv2) H1) as [p ?].
  destruct (H0 (Ctx_OPB2 _ ctx (tvs1 p)) H2) as [p' ?].
  by exists (max p p');
   apply (ParamLeV_ctx_chain (CClosV_chain Hc) (le_max_l _ _) (Ctx_OPB1 _ _ _));
   apply (ParamLeV_ctx_chain (CClosV_chain Hc0) (le_max_r _ _) (Ctx_OPB2 _ _ _)).
- by destruct (H (Ctx_FST _ ctx) H0) as [p ?]; exists p.
- by destruct (H (Ctx_SND _ ctx) H0) as [p ?]; exists p.
- destruct (H (Ctx_CASE1 ctx tv1 tv2) H2) as [p ?].
  destruct (H0 (Ctx_CASE2 ctx (tvs p) tv2) H3) as [p' ?].
  destruct (H1 (Ctx_CASE3 ctx (tvs p) (tvs1 p')) H4) as [p'' ?].
  have lep' : p' <= max p (max p' p''); [ |
   have lep'' : p'' <= max p (max p' p'') ];
    try by apply le_trans with (max p' p''); auto with arith.
   by exists (max p (max p' p''));
    apply (ParamLeV_ctx_chain (CClosV_chain Hc)(le_max_l _ _)(Ctx_CASE1 _ _ _));
    apply (ParamLeC_ctx_chain (CClosC_chain Hc0) lep' (Ctx_CASE2 _ _ _));
    apply (ParamLeC_ctx_chain (CClosC_chain Hc1) lep'' (Ctx_CASE3 _ _ _)).
- destruct (H (Ctx_LET1 ctx tv2) H1) as [p ?].
  destruct (H0 (Ctx_LET2 ctx (tvs1 p)) H2) as [p' ?].
  by exists (max p p');
   apply (ParamLeC_ctx_chain (CClosC_chain Hc) (le_max_l _ _) (Ctx_LET1 _ _));
   apply (ParamLeC_ctx_chain (CClosC_chain Hc0) (le_max_r _ _) (Ctx_LET2 _ _)).
- destruct (H (Ctx_APP1 ctx tv2) H1) as [p ?].
  destruct (H0 (Ctx_APP2 ctx (tvs1 p)) H2) as [p' ?].
  by exists (max p p');
   apply (ParamLeV_ctx_chain (CClosV_chain Hc) (le_max_l _ _) (Ctx_APP1 _ _));
   apply (ParamLeV_ctx_chain (CClosV_chain Hc0) (le_max_r _ _) (Ctx_APP2 _ _)).
- by destruct (H (Ctx_TAPP _ _ ctx) H0) as [p ?]; exists p.
- destruct (H (Ctx_UNPK1 ctx tv2) H1) as [p ?].
  destruct (H0 (Ctx_UNPK2 ctx (tvs1 p)) H2) as [p' ?].
  by exists (max p p');
   apply (ParamLeV_ctx_chain (CClosV_chain Hc) (le_max_l _ _) (Ctx_UNPK1 _ _));
   apply (ParamLeC_ctx_chain (CClosC_chain Hc0) (le_max_r _ _) (Ctx_UNPK2 _ _)).
- by apply (SLeC_ctx_adeq Hs _ H).
Qed.

Definition CmptClos : AdqCmpt (@ParamLeV ParamLe) (@ParamLeC ParamLe).
apply (@mkAdqCmpt _ _ CmptClosV CmptClosC); 
  try (by apply CClos_chain); try (by constructor).
- by red; intros; apply (proj2 CClos_ctx_adeq _ _ _ _ _ H Ctx_ID).
Defined.

End Cmpt_Closure.

(***
   Largest precongruence between terms.
 ***)

Definition IdCmpt : AdqCmpt IdV IdC.
apply (@mkAdqCmpt _ _ SIdV SIdC); repeat red; intros;
  try red in H; try red in H0; try red in H1; subst; auto;
  try (by extension p'; destruct (le_lt_dec p' p); 
          [ | symmetry]; first [apply IdV_eq|apply IdC_eq]; apply H; omega);
  try (by extension p'; unfold cstr in *; f_equal;
          args2 extension T0 var0; apply (f_app p' (H T0 var0))).
- by destruct H0; exists 0; eexists; Apply.
Defined.

Lemma IdCmptV_trans : STRelV_trans (AdqCmptV IdCmpt).
Proof. by red; intros; specialize (H0 0); repeat red in H, H0 |- *; subst. Qed.

Lemma IdCmptC_trans : STRelC_trans (AdqCmptC IdCmpt).
Proof. by red; intros; specialize (H0 0); repeat red in H, H0 |- *; subst. Qed.

Definition BigLe : AdqCmpt IdV IdC
  := UnionCmpt IdCmpt IdCmptV_trans IdCmptC_trans.

Definition BigLeV : TRelV := fun _ _ _ tv tv' => AdqCmptV BigLe tv (cstr tv').
Definition BigLeC : TRelC := fun _ _ _ tv tv' => AdqCmptC BigLe tv (cstr tv').
 
(***
   Contextual precongruence between a term and a chain of terms
 ***)

Definition SBigLe : AdqCmpt BigLeV BigLeC :=
  UnionCmpt BigLe (@UnionCmptV_trans IdCmpt) (@UnionCmptC_trans IdCmpt).

Definition SBigLeV : STRelV := AdqCmptV SBigLe.
Definition SBigLeC : STRelC := AdqCmptC SBigLe.

(***
   Equivalence of Contextual Equivalence and the Largest precongruence
 ***)

Lemma AdqCmpt_is_closed_under_context: forall ParamLe (rel : AdqCmptP ParamLe),
  (forall u (E : Env u) T (ctx : CtxV E T) tv tvs, 
    AdqCmptV rel tv tvs -> 
    AdqCmptC rel (pluginV ctx tv) (fun p => pluginV ctx (tvs p))) /\
  (forall u (E : Env u) T (ctx : CtxC E T) tv tvs,
    AdqCmptC rel tv tvs -> 
    AdqCmptC rel (pluginC ctx tv) (fun p => pluginC ctx (tvs p))).
Proof.
move => ParamLe rel; apply CtxVC_ind; intros;
 try (by simpl; rewrite <- eta_expansion);
 try (by apply H; apply_AdqCmpt; auto; 
        solve [apply AdqCmptV_cstr_refl|apply AdqCmptC_cstr_refl]).
Qed.

Definition CtxLeCClos : AdqCmptP IdCmpt.
apply (@CmptClos IdCmpt (to_STRelV CtxLeV) (to_STRelC CtxLeC)); red; intros;
  try (by repeat red in H; unfold cstr in *; rewrite (f_app 0 (H _ _ H0)));
  try (by destruct H as [? [? ?]]; subst);
  try (by destruct H as [? [? ?]]; subst; exists 0; Apply). 
Defined.

Lemma BigLeV_le_CtxLeV: subrel_TRelV BigLeV CtxLeV.
Proof.
do 2 red; intros; destruct H as [rel ?].
by apply (AdqCmpt_adeq rel _ 
  (proj1 (AdqCmpt_is_closed_under_context _) _ _ _ _ _ _ H)).
Qed.

Lemma BigLeC_le_CtxLeC: subrel_TRelC BigLeC CtxLeC.
Proof.
do 2 red; intros; destruct H as [rel ?].
by apply (AdqCmpt_adeq rel _ 
  (proj2 (AdqCmpt_is_closed_under_context _) _ _ _ _ _ _ H)).
Qed.

Lemma CtxLeV_le_BigLeV: subrel_TRelV CtxLeV BigLeV.
Proof. by do 2 red; intros; exists CtxLeCClos; constructor; exists tv'. Qed.

Lemma CtxLeC_le_BigLeC: subrel_TRelC CtxLeC BigLeC.
Proof. by do 2 red; intros; exists CtxLeCClos; constructor; exists tv'. Qed.

Definition SCtxLeCClos : AdqCmptP BigLe.
apply (@CmptClos BigLe SCtxLeV SCtxLeC); red; intros; try by Apply.
- by apply BigLeV_le_CtxLeV; apply H.
- by apply BigLeC_le_CtxLeC; apply H.
- by destruct H; red; intros; apply CtxLeV_le_BigLeV; Apply.
- by destruct H; red; intros; apply CtxLeC_le_BigLeC; Apply.
Defined.

Lemma SBigLeV_le_SCtxLeV: subrel_STRelV SBigLeV SCtxLeV.
Proof.
do 2 red; intros; destruct H as [rel ?]; split; red; intros.
- by apply BigLeV_le_CtxLeV; eapply AdqCmptV_chain; Apply.
- eapply (AdqCmpt_adeq rel); last by Apply.
  by apply (AdqCmpt_is_closed_under_context _); Apply.
Qed.

Lemma SBigLeC_le_SCtxLeC: subrel_STRelC SBigLeC SCtxLeC.
Proof.
do 2 red; intros; destruct H as [rel ?]; split; red; intros.
- by apply BigLeC_le_CtxLeC; eapply AdqCmptC_chain; Apply.
- eapply (AdqCmpt_adeq rel); last by Apply.
  by apply (AdqCmpt_is_closed_under_context _); Apply.
Qed.

Lemma SCtxLeV_le_SBigLeV: subrel_STRelV SCtxLeV SBigLeV.
Proof. by do 2 red; intros; exists SCtxLeCClos; constructor. Qed.

Lemma SCtxLeC_le_SBigLeC: subrel_STRelC SCtxLeC SBigLeC.
Proof. by do 2 red; intros; exists SCtxLeCClos; constructor. Qed.

(****
   SBigLe = SCtxLe is an instance of ObsLe
 ****)

Definition SSVal u (x: Env u * Ty u) : Type := 
  match x with (env, ty) => stream (Val u env ty) end.
Notation   PSVal := (_,_) (only parsing).
Implicit Arguments SSVal [[u]].

Definition SSExp u (x: Env u * Ty u) : Type := 
  match x with (env, ty) => stream (Exp u env ty) end.
Notation   PSExp := (_,_) (only parsing).
Implicit Arguments SSExp [[u]].

Lemma BigLeV_eq_SBigLeV_cstr: forall u E T (tv tv': Val u E T),
  BigLeV tv tv' <-> SBigLeV tv (cstr tv').
Proof. 
split; intros.
- apply SCtxLeV_le_SBigLeV; apply BigLeV_le_CtxLeV in H.
  by split; [red; intros | exists 0; auto].
- apply CtxLeV_le_BigLeV; apply SBigLeV_le_SCtxLeV in H.
  by destruct H as [Hcha Hle]; move => ctx Hterm; destruct (Hle _ Hterm); Apply.
Qed.

Lemma BigLeC_eq_SBigLeC_cstr: forall u E T (tv tv': Exp u E T),
  BigLeC tv tv' <-> SBigLeC tv (cstr tv').
Proof. 
split; intros.
- apply SCtxLeC_le_SBigLeC; apply BigLeC_le_CtxLeC in H.
  by split; [red; intros | exists 0; auto].
- apply CtxLeC_le_BigLeC; apply SBigLeC_le_SCtxLeC in H.
  by destruct H as [Hcha Hle]; move => ctx Hterm; destruct (Hle _ Hterm); Apply.
Qed.

Lemma BigLe_app_beta: TRel_app_beta BigLeC.
Proof. by split; apply CtxLeC_le_BigLeC; apply (CtxLe_app_beta _ _). Qed.

Lemma SBigLe_app_beta: TRel_app_beta (to_TRelC SBigLeC).
Proof. 
by split; apply (BigLeC_eq_SBigLeC_cstr _ _); apply (BigLe_app_beta _ _).
Qed.

Lemma BigLe_tyapp_beta: TRel_tyapp_beta BigLeC.
Proof. by split; apply CtxLeC_le_BigLeC; apply (CtxLe_tyapp_beta _ _ _). Qed.

Lemma SBigLe_tyapp_beta: TRel_tyapp_beta (to_TRelC SBigLeC).
Proof. 
by split; apply (BigLeC_eq_SBigLeC_cstr _ _); apply (BigLe_tyapp_beta _ _ _).
Qed.

Lemma SBigLeC_value : forall u E T (tv: Val u E T) tvs,
  SBigLeC (^tv) (fun p => ^tvs p) -> SBigLeV tv tvs.
Proof. 
by intros; apply SCtxLeV_le_SBigLeV, SCtxLeC_value, SBigLeC_le_SCtxLeC.
Qed.

Lemma BigLeC_value : forall u E T (tv: Val u E T) tv',
  BigLeC (^tv) (^tv') -> BigLeV tv tv'.
Proof.
intros; apply (BigLeV_eq_SBigLeV_cstr _ _), SBigLeC_value.
by apply (proj1 (BigLeC_eq_SBigLeC_cstr _ _)) in H.
Qed.

Lemma BigLeV_refl : TRelV_refl BigLeV.
Proof. by red; intros; apply CtxLeV_le_BigLeV. Qed.

Lemma BigLeC_refl : TRelC_refl BigLeC.
Proof. by red; intros; apply CtxLeC_le_BigLeC. Qed.

Hint Resolve BigLeV_refl BigLeC_refl.

Lemma BigLeV_trans : TRelV_trans BigLeV.
Proof. by red; intros; eapply (@UnionCmptV_trans IdCmpt); intros; Apply. Qed.

Lemma BigLeC_trans : TRelC_trans BigLeC.
Proof. by red; intros; eapply (@UnionCmptC_trans IdCmpt); intros; Apply. Qed.

Lemma SBigLeV_trans : STRelV_trans SBigLeV.
Proof. by apply (@UnionCmptV_trans BigLe). Qed.

Lemma SBigLeC_trans : STRelC_trans SBigLeC.
Proof. by apply (@UnionCmptC_trans BigLe). Qed.

Lemma BigLeV_Sub: TRelV_Sub BigLeV.
Proof.
red; induction E; intros.
- replace Sub with (RTS (@nilMap _ _ E')) by Mapext_destr.
  replace Sub' with (RTS (@nilMap _ _ E')) by Mapext_destr.
  by apply_AdqCmpt.
- apply BigLeC_value; apply BigLeC_trans with 
    (APP (SubVal (tlMap Sub) (LAM (^tv))) (hdMap Sub)).
  - eabs (^SubVal Sub tv); first by apply (BigLe_app_beta _ _).
    by unfold_SubVE; simpl_Sub; rewrite <- consMapInv.
  - apply BigLeC_trans with 
      (APP (SubVal (tlMap Sub') (LAM (^tv'))) (hdMap Sub')).
    - apply_AdqCmpt; last by apply_AdqCmpt; Apply.
      by apply IHE; [intros; Apply | apply_AdqCmpt].
    - eapply BigLeC_trans; first by eapply (BigLe_app_beta _ _).
      by unfold_SubVE; simpl_Sub; rewrite <- consMapInv.
Qed.

Lemma BigLeC_Sub: TRelC_Sub BigLeV BigLeC.
Proof.
red; intros; apply BigLeC_trans with 
  (APP (SubVal Sub (LAM (SubExp (RTS (shRen UNIT)) tv))) VUNIT).
- unfold_SubVE; eabs (SubExp Sub tv); first by apply (BigLe_app_beta _ _).
  by unfold_SubVE; simpl_Sub; rewrite <- consMapInv.
- apply BigLeC_trans with 
    (APP (SubVal Sub' (LAM (SubExp (RTS (shRen UNIT)) tv'))) VUNIT).
  - apply_AdqCmpt; last by apply_AdqCmpt; Apply.
    by apply BigLeV_Sub; [intros; Apply | apply_AdqCmpt].
  - unfold_SubVE; eapply BigLeC_trans; first by eapply (BigLe_app_beta _ _).
    by unfold_SubVE; simpl_Sub.
Qed.

Lemma SBigLeV_Sub: STRelV_Sub SBigLeV.
Proof. 
red; induction E; intros.
- replace Sub with (RTS (@nilMap _ _ E')) by Mapext_destr.
  eapply eq_ind; first by apply (AdqCmptV_Ren SBigLe _ _ _ H0).
  by extension p; replace (Subs p) with (RTS (@nilMap _ _ E')) by Mapext_destr.
- apply SBigLeC_value; apply SBigLeC_trans with 
    (cstr (APP (SubVal (tlMap Sub) (LAM (^tv))) (hdMap Sub))).
  - eabs (^SubVal Sub tv); first by apply (SBigLe_app_beta _ _).
    by unfold_SubVE; simpl_Sub; rewrite <- consMapInv.
  - unfold cstr; move => _; apply SBigLeC_trans with 
      (fun p => APP (SubVal (tlMap (Subs p)) (LAM (^tvs p))) (hdMap (Subs p))).
    - apply_AdqCmpt; last by apply_AdqCmpt; Apply.
      by apply IHE; [intros; Apply | apply_AdqCmpt].
    - move => p; unfold_SubVE.
      eapply SBigLeC_trans; first by apply (SBigLe_app_beta _ _).
      unfold cstr; move => _; unfold_SubVE; simpl_Sub; rewrite <- consMapInv.
      apply_AdqCmpt.
      apply (@AdqCmptV_refl _ _ SBigLe _ _ _ (fun p=>SubVal (Subs p) (tvs p))).
      red; intros; apply BigLeV_Sub; intros. 
      - by apply (AdqCmptV_chain SBigLe _ _ (H _ var)).
      - by apply (AdqCmptV_chain SBigLe _ _ H0).
Qed.

Lemma SBigLeC_Sub: STRelC_Sub SBigLeV SBigLeC.
Proof. 
red; intros; apply SBigLeC_trans with 
  (cstr (APP (SubVal Sub (LAM (SubExp (RTS (shRen UNIT)) tv))) VUNIT)).
- eabs (SubExp Sub tv); first by apply (SBigLe_app_beta _ _).
  by unfold_SubVE; simpl_Sub; rewrite <- consMapInv.
- unfold cstr; move => _; apply SBigLeC_trans with 
    (fun p => APP (SubVal (Subs p) (LAM (SubExp (RTS (shRen UNIT)) (tvs p))))
      VUNIT).
  - apply_AdqCmpt; last by apply_AdqCmpt; Apply.
    by apply SBigLeV_Sub; [intros; Apply | apply_AdqCmpt].
  - move => p; unfold_SubVE.
    eapply SBigLeC_trans; first by apply (SBigLe_app_beta _ _).
    unfold cstr; move => _; unfold_SubVE; simpl_Sub.
    apply (@AdqCmptC_refl _ _ SBigLe _ _ _ (fun p=>SubExp (Subs p) (tvs p))).
    red; intros; apply BigLeC_Sub; intros. 
    - by apply (AdqCmptV_chain SBigLe _ _ (H _ var)).
    - by apply (AdqCmptC_chain SBigLe _ _ H0).
Qed.

Lemma BigLeV_le_C1: forall u w (sub:subT (S u) w) (E: Env w) T,
  (E, subTy [|hdmap sub|] (subTy (subL (tlmap sub)) T)) = (E, subTy sub T).
Proof. by intros; simpl_sub; repeat f_equal; mapext_destr. Qed.
Implicit Arguments BigLeV_le_C1 [[u] [w] [sub] [E] [T]].

Lemma BigLeV_sub: TRelV_sub BigLeV.
Proof.
red; induction u; intros; first
  by replace sub with (rts (@nilmap _ w)); [ apply_AdqCmpt | mapext_destr].
move : w sub T tv tv' H; induction E; intros.
- apply BigLeC_value; apply BigLeC_trans with (<<SExp,PExp#BigLeV_le_C1>>
    TAPP (subVal (tlmap sub) (TLAM (env:=nil) tv)) (hdmap sub)).
  - Hrefl (subVal sub tv : SVal PVal); Helimc C; intros; unfold_SubVE.
    eabs(<<SVal,PVal#C>>subVal sub tv); first by apply (BigLe_tyapp_beta [] _ _).
    apply Heq_eq; Hsimplc; rewrite (Hrw (subVal_ss tv _ _)); Helimc.
    by simpl_sub; rewrite <- consmapInv.
  - apply BigLeC_trans with (<< SExp # BigLeV_le_C1 >>
      TAPP (subVal (tlmap sub) (TLAM (env:=[]) tv')) (hdmap sub)).
    - Hsimplc; apply_AdqCmpt; fold_sub.
      apply (IHu _ (tlmap sub) _ _ _ (TLAM (env:=[]) tv')).
      by apply_AdqCmpt.
    - unfold_SubVE; Hsimplc; intros; eapply BigLeC_trans. 
      - Hhide_evars; Hrefl (hev0 : SExp PExp); Helimc; Hshow_evars; intros.
        eapply eq_ind; first by apply (BigLe_tyapp_beta [] _ _).
        apply (Heq_eq SExp PExp); Hhide_evars; Hsimplc.
        rewrite (Hrw (subVal_ss tv' _ _)); Hsimplc.
        by simpl_sub; rewrite <- consmapInv; Hshow_evars.
      - by [].
- apply BigLeC_value; apply BigLeC_trans with 
    (APP (SubVal (shSub (subTy sub a)) (subVal sub (LAM (^tv)))) (#V 0)).
  - unfold_SubVE; eapply BigLeC_trans; last by apply (BigLe_app_beta _ _).
    by unfold_SubVE; simpl_Sub; rewrite consMap_ZVAR_shSub; simpl_Sub.
  - apply BigLeC_trans with
      (APP (SubVal (shSub (subTy sub a)) (subVal sub (LAM (^tv')))) (#V 0)).
    - by apply_AdqCmpt; [apply IHE; apply_AdqCmpt | apply BigLeV_refl].
    - unfold_SubVE; eapply BigLeC_trans; first by apply (BigLe_app_beta _ _).
      by unfold_SubVE; simpl_Sub; rewrite consMap_ZVAR_shSub; simpl_Sub.
Qed.

Lemma BigLeC_sub: TRelC_sub BigLeC.
Proof.
red; intros; apply BigLeC_trans with 
  (APP (subVal sub (LAM (SubExp (RTS (shRen UNIT)) tv))) VUNIT).
- unfold_SubVE; eabs (subExp sub tv); first by apply (BigLe_app_beta _ _).
  by do 2 simpl_Sub.
- apply BigLeC_trans with 
    (APP (subVal sub (LAM (SubExp (RTS (shRen UNIT)) tv'))) VUNIT).
  - apply_AdqCmpt; fold_sub; last by apply_AdqCmpt; Apply. 
    apply (@BigLeV_sub _ _ sub _ _ _ (LAM (SubExp (RTS (shRen UNIT)) tv'))).
    by apply_AdqCmpt.
  - unfold_SubVE; eapply BigLeC_trans; first by eapply (BigLe_app_beta _ _).
    by unfold_SubVE; do 2 simpl_Sub.
Qed.

Lemma SBigLeV_sub: STRelV_sub SBigLeV.
Proof.
red; induction u; intros; first
  by replace sub with (rts (@nilmap _ w)); [ apply_AdqCmpt | mapext_destr].
move : w sub T tv tvs H; induction E; intros.
- apply SBigLeC_value; apply SBigLeC_trans with (cstr(<<SExp,PExp#BigLeV_le_C1>>
    TAPP (subVal (tlmap sub) (TLAM (env:=nil) tv)) (hdmap sub))).
  - Hrefl (subVal sub tv : SVal PVal); Helimc C; intros; unfold_SubVE.
    eabs(<<SVal,PVal#C>>subVal sub tv); first by apply(SBigLe_tyapp_beta [] _ _).
    apply Heq_eq; Hsimplc; rewrite (Hrw (subVal_ss tv _ _)); Helimc.
    by simpl_sub; rewrite <- consmapInv.
  - move => p; apply SBigLeC_trans with (fun p => << SExp # BigLeV_le_C1 >>
      TAPP (subVal (tlmap sub) (TLAM (env:=[]) (tvs p))) (hdmap sub)).
    - Hsimplc; apply_AdqCmpt; fold_sub.
      apply (IHu _ (tlmap sub) _ _ _ (fun p => TLAM (env:=[]) (tvs p))).
      by apply_AdqCmpt.
    - unfold_SubVE; Hsimplc; intros; eapply SBigLeC_trans. 
      - Hhide_evars; Hrefl (hev0 : SSExp PSExp); Helimc; Hshow_evars; intros.
        eapply eq_ind; first by apply (SBigLe_tyapp_beta [] _ _).
        apply (Heq_eq SSExp PSExp); Hhide_evars; Hsimplc.
        rewrite (Hrw (subVal_ss (tvs p0) _ _)); Hsimplc.
        by simpl_sub; rewrite <- consmapInv; Hshow_evars.
      - unfold cstr; move => _; apply_AdqCmpt.
        apply (@AdqCmptV_refl _ _ SBigLe _ _ _ (fun p => subVal sub (tvs p))).
        red; intros; apply BigLeV_sub; intros. 
        by apply (AdqCmptV_chain SBigLe _ _ H).
- apply SBigLeC_value; apply SBigLeC_trans with 
    (cstr (APP (SubVal (shSub (subTy sub a)) (subVal sub (LAM (^tv)))) (#V 0))).
  - unfold_SubVE; eapply SBigLeC_trans with (cstr _); 
      last by intros; apply(SBigLe_app_beta _ _).
    unfold_SubVE; simpl_Sub; rewrite consMap_ZVAR_shSub; simpl_Sub.
    by apply AdqCmptC_cstr_refl.
  - unfold cstr; move => _; apply SBigLeC_trans with (fun p =>
      (APP (SubVal (shSub (subTy sub a)) (subVal sub (LAM (^tvs p)))) (#V 0))).
    - by apply_AdqCmpt; [apply IHE; apply_AdqCmpt | apply AdqCmptV_cstr_refl].
    - intros; unfold_SubVE.
      eapply SBigLeC_trans; first by apply (SBigLe_app_beta _ _).
      unfold cstr; move => _.
      unfold_SubVE; simpl_Sub; rewrite consMap_ZVAR_shSub; simpl_Sub.
      apply_AdqCmpt.
      apply (@AdqCmptV_refl _ _ SBigLe _ _ _ (fun p => subVal sub (tvs p))).
      red; intros; apply BigLeV_sub; intros. 
      by apply (AdqCmptV_chain SBigLe _ _ H).
Qed.

Lemma SBigLeC_sub: STRelC_sub SBigLeC.
Proof. 
red; intros; apply SBigLeC_trans with 
  (cstr (APP (subVal sub (LAM (SubExp (RTS (shRen UNIT)) tv))) VUNIT)).
- eabs (subExp sub tv); first by apply (SBigLe_app_beta _ _).
  by unfold_SubVE; do 2 simpl_Sub. 
- unfold cstr; move => _; apply SBigLeC_trans with 
    (fun p => APP (subVal sub (LAM (SubExp (RTS (shRen UNIT)) (tvs p))))
      VUNIT).
  - apply_AdqCmpt; last by apply_AdqCmpt; Apply.
    apply (@SBigLeV_sub _ _ sub _ _ _ 
      (fun p => LAM (SubExp (RTS (shRen UNIT)) (tvs p)))).
    by apply_AdqCmpt.
  - move => p; unfold_SubVE.
    eapply SBigLeC_trans; first by apply (SBigLe_app_beta _ _).
    unfold cstr; move => _; unfold_SubVE; do 2 simpl_Sub.
    apply (@AdqCmptC_refl _ _ SBigLe _ _ _ (fun p => subExp sub (tvs p))).
    red; intros; apply BigLeC_sub; intros. 
    - by apply (AdqCmptC_chain SBigLe _ _ H).
Qed.

(***
   (SBigLe = SCtxLe) forms a module instance of ObsLeSig.
 ***)

Module ContextualApproximation : ObsLeSig.

Definition ole : ObsLe.
apply (@mkObsLe SBigLeV SBigLeC); try apply_AdqCmpt; auto.
- do 2 red; apply AdqCmptV_cstr_refl.
- do 2 red; apply AdqCmptC_cstr_refl.
- by red; intros; split; red; 
     apply (BigLeC_eq_SBigLeC_cstr _ _), CtxLeC_le_BigLeC, (CtxLe_beta _ _).
- do 2 red; intros; red in H.
  by apply (BigLeV_eq_SBigLeV_cstr _ _), (AdqCmptV_chain _ _ _ H).
- do 2 red; intros; red in H.
  by apply (BigLeC_eq_SBigLeC_cstr _ _), (AdqCmptC_chain _ _ _ H).
- red; intros; apply_AdqCmpt; red; intros.
  by apply (BigLeV_eq_SBigLeV_cstr _ _); Apply.
- red; intros; apply_AdqCmpt; red; intros.
  by apply (BigLeC_eq_SBigLeC_cstr _ _); Apply.
- by apply SBigLeV_trans.
- by apply SBigLeC_trans.
- by apply SBigLeV_sub.
- by apply SBigLeC_sub.
- by apply SBigLeV_Sub.
- by apply SBigLeC_Sub.
- do 2 red; intros; apply SCtxLeV_le_SBigLeV; split; repeat red; intros.
  - eapply CtxLe_recn_monotone; Apply.
  - by apply CtxLe_unwinding.
Defined.

End ContextualApproximation.
