(*******************************************************************************
   Proof of optimized list module for
   Realizability and Compositional Compiler Correctness 
      for a Polymorphic Language
   N Benton & C-K Hur
   April 2010
*******************************************************************************)

Require Import Arith List Program FunctionalExtensionality Heq.
Require Import util.
Require Import ssreflect ssrbool Heq.
Require Import cesd sysf sysfopsem sysfopsem_reflect logrel compcorrect.

Set Implicit Arguments.

(******
   Polymorphic List
 ******)

Record NatPairing :=
  mkNatPairing {
    pair_nat: nat -> nat -> nat;
    fst_nat: nat -> nat;
    snd_nat: nat -> nat;

    pair_nat_pos: forall n m, pair_nat n m > 0;
    fst_nat_eq: forall n m, fst_nat (pair_nat n m) = n;
    snd_nat_eq: forall n m, snd_nat (pair_nat n m) = m;

    PairNat : Code;
    SplitNat: Code;

    PairNat_obs: forall cesd n m,
      exists k,
        kstepreduce k (mergeComp (PairNat, [VN m; VN n]) cesd) 
        (mergeComp (nil, [VN (pair_nat n m)]) cesd);

    SplitNat_obs: forall cesd n,
      exists k, 
        kstepreduce k (mergeComp (SplitNat, [VN n]) cesd) 
        (mergeComp (nil, [VP (VN (fst_nat n)) (VN (snd_nat n))]) cesd)
  }.

Module Type NatPairingSig.

  Parameter NPR: NatPairing.

End NatPairingSig.

Module EgPolList (ObsLeMod: ObsLeSig) (NatPairMod : NatPairingSig).

Module ObsLeInstance := ObsLeMod.
Module LogRelInstance := LogicalRelationFunctor (ObsLeInstance).
Module CompCorrect := CompilerCorrectFunctor (ObsLeInstance).
Import LogRelInstance.
Import CompCorrect.
Import ObsLeMod.
Import NatPairMod.

(*
   Source Code
*)

Definition Option u (T : Ty u) : Ty u := SUM UNIT T.

Definition NONE u E T : Val u E (Option T) := INL _ VUNIT.
Definition SOME u E T (a: Val u E T) : Val u E (Option T) := INR _ a.

(* Internal Implementation of List *)
Definition List u (T: Ty u) : Ty u := 
  ALL (#TV 0 *** ((UP T) *** #TV 0 --> #TV 0) --> #TV 0).

(* NIL :=
   TLAM X. LAM (nilcons). FST nilcons
*)  
Definition NIL u E T : Val u E (List T) := (TLAM (LAM (FST (#V 0)))). 

(* CONS a tl := 
   TLAM X. lam nilcons.
     LET tlX := tl [X] IN
       LET tlXnc := APP tlX nilcons IN
         LET cons := SND nilcons IN
           APP cons (a, tlXnc) 
*)
Lemma CONS_C1: forall u (E:Env (S u)) (T:Ty u), 
  (E, UP (List T)) = (E, List (UP T)).
Proof. by intros; unfold List, UP; simpl_sub. Qed.
Implicit Arguments CONS_C1 [u E T].

Lemma CONS_C2: forall u (E:Env u) (T t: Ty u),
  (E, subTy [|t|] ((#TV 0) *** ((UP T) *** (#TV 0) --> (#TV 0)) --> (#TV 0))) = 
  (E, t *** (T *** t --> t) --> t).
Proof. by intros; unfold List, UP; simpl_sub. Qed.
Implicit Arguments CONS_C2 [u E T t].

Program Definition CONS u E T (a: Val u E T) (tl: Val u E (List T)) 
  : Val u E (List T) 
  := TLAM 
       (LAM 
         (LET (TAPP (UPVal [_] (<< SVal,PVal # CONS_C1 >> upVal tl)) (#TV 0))
           (LET (APP (<< SVal,PVal # CONS_C2 >> #V 0) (#V 1))         
             (LET (SND (#V 2))                                        
               (APP (#V 0) (PAIR (UPVal [_;_;_;_] (upVal a)) (#V 1))))))).

(* SPLIT l :=
   LET l' := l [Option (T**List T)] IN
     APP l' (NONE, 
             LAM hdtl. LET hd := FST hdtl IN LET tl := SND hdtl IN
               CASE tl WITH
               | NONE => SOME (hd, NIL)
               | SOME hdtl' => LET hd' := FST hdtl' IN 
                               LET tl' := SND hdtl' IN 
                                 SOME (hd, CONS hd' tl'))
*)
Lemma SPLIT_C1: forall u (E:Env u) (T: Ty u),
  (E, T *** Option (T *** List T) --> Option (T *** List T)) =
  (E, subTy [|Option (T *** List T)|] (UP T) *** 
    Option (T ***  List T) --> Option (T ***  List T)).
Proof. by intros; unfold List, UP; simpl_sub. Qed.
Implicit Arguments SPLIT_C1 [u E T].

Program Definition SPLIT u E T (l: Val u E (List T)) 
  : Exp u E (Option (T *** List T)) 
  :=
  LET (TAPP l (Option (T *** List T)))
   (APP (#V 0) (PAIR (NONE _ _) 
               (<< SVal,PVal # SPLIT_C1 >> 
                LAM (LET (FST (#V 0)) (LET (SND (#V 1))
                      (CASE (#V 0)
                        (VAL (SOME (PAIR (#V 2) (NIL _ _))))
                        (LET (FST (#V 0)) (LET (SND (#V 1)) 
                         (VAL (SOME (PAIR (#V 4) (CONS (#V 1) (#V 0))))))))))))).

(*
   List Module
*)

Definition NilTerm : Val 1 nil (List (#TV 0)) := NIL _ _.

(* ConsTerm :=
   LAM hdtl. LET hd := FST hdtl IN LET tl := SND hdtl IN CONS hd tl *)
Program Definition ConsTerm : 
  Val 1 nil (#TV 0 *** List (#TV 0) --> List (#TV 0)) 
  := 
  LAM (LET (FST (#V 0)) (LET (SND (#V 1)) (VAL (CONS (#V 1) (#V 0))))).

(* SplitTerm :=
   LAM l. SPLIT l *)
Definition SplitTerm : 
  Val 1 nil (List (#TV 0) --> Option (#TV 0 *** List (#TV 0))) 
  := LAM (SPLIT (#V 0)).

Definition ListSig := 
  ALL (u:=0) (EXIST (
    #TV 0 *** 
    (#TV 1 *** #TV 0 --> #TV 0) *** 
    (#TV 0 --> Option (#TV 1 *** #TV 0)))).

Program Definition ListModule : CVal ListSig := 
  TLAM (PACK (PAIR NilTerm (PAIR ConsTerm SplitTerm))).

(*****
   Properties 
 *****)

(* Definitions and Tactics *)

Definition Sirel (ty: Ty 0) : Type := irel MValue (CVal ty).
Definition Srel (ty: Ty 0) : Type := rel MValue (CVal ty).

Definition ListR u (T: Ty u) : Ty u := 
  ALL (#TV 0 *** ((renTy shren T) *** #TV 0 --> #TV 0) --> #TV 0).

Lemma List_eq_ListR: forall u (T:Ty u), List T = ListR T.
Proof. by intros; unfold List, ListR, UP; simpl_sub. Defined.

Ltac Lunfold := unfold Option, List, ListR, UPVal, upVal, UP.
Ltac Lsimpl_sub := Lunfold; simpl_sub.
Ltac Lsimpl_Sub := Lunfold; simpl_Sub.

Ltac Lhide := 
  remember NilTerm; remember ConsTerm; remember SplitTerm; 
  remember NIL; remember CONS; remember SPLIT.
Ltac Lshow := subst.

(*
   Machine Code
*)

(* Nil *)
Definition NilCode : Code (* Type: List T *)
  := [PushN 0].

(* Cons *)
Definition ConsCode : Code (* Type: T ** List T --> List T *)
  :=
  [PushC [
    Push 0; ProjFst;
    IsNum;
    Sel [Push 0; ProjSnd; IsNum; Join] [Push 0; ProjSnd; PushN 0; Join];
    Sel (PairNat NPR ++ [Join]) [MkPair; Join];
    Ret
  ]].
  
(* Split *)
Definition SplitCode : Code (* Type: List T --> Option (T ** List T) *)
  :=
  [PushC [
    Push 0;
    Sel [ PushN 0; 
          Push 0; IsNum; Sel (SplitNat NPR ++ [Join]) [Join]; 
          MkPair; Join ] 
        [ PushN 1; PushN 1; MkPair; Join ];
    Ret
  ]].

(* ListModule *)

Definition ListModuleCode : Code 
  :=
  NilCode ++ ConsCode ++ SplitCode ++ [MkPair] ++ [MkPair].

(* list representations *)

Fixpoint rel_list (A B:Type) (r: rel A B) (al: list A) (bl: list B) : Prop :=
  match (al, bl) with
    | (nil, nil) => True
    | (a::al', b::bl') => r a b /\ rel_list r al' bl'
    | _ => False
  end.

Fixpoint encode_list_cesd (vl: list MValue) : MValue :=
  match vl with
    | nil => VN 0
    | v :: vl' => 
      match (v, encode_list_cesd vl') with
        | (VN n1, VN n2) => VN (pair_nat NPR n1 n2)
        | (v1, v2) => VP v1 v2
      end
  end.

Fixpoint encode_list_src (t: Ty 0) (tvl: list (CVal t)) : CVal (List t) :=
  match tvl with
    | nil => NIL _ _
    | tv :: tvl' => CONS tv (encode_list_src tvl')
  end.

Definition relList_le (t: Ty 0) (tr: irel MValue (CVal t)) 
  : irel MValue (CVal (List t))
  := fun i v tv =>
    exists vl, exists tvl, 
      rel_list (tr i) vl tvl /\
      v = encode_list_cesd vl /\
      tv = encode_list_src tvl.

Definition relList_ge (t: Ty 0) (tr: rel MValue (CVal t))
  : rel MValue (CVal (List t))
  := fun v tv =>
    exists vl, exists tvl, 
      rel_list tr vl tvl /\
      v = encode_list_cesd vl /\
      tv = encode_list_src tvl.

Lemma encode_list_cesd_nil: forall vl,
  encode_list_cesd vl = VN 0 -> vl = nil.
Proof.
intros. destruct vl; first by [].
destruct m; inversion H.
destruct (encode_list_cesd vl); inversion H1.
assert (H3 := pair_nat_pos NPR n n0).
rewrite H2 in H3. inversion H3.
Qed.

Lemma encode_list_cesd_num: forall tl hd,
  MValue_is_num (encode_list_cesd (hd :: tl)) ->
  exists n0, exists n1,
    hd = VN n0 /\ encode_list_cesd tl = VN n1.
Proof.
intros; case_eq (encode_list_cesd (hd::tl)); 
  (intros; rewrite H0 in H; inversion H).
destruct hd; inversion H0.
case_eq (encode_list_cesd tl); (intros; rewrite H1 in H2; inversion H2).
eexists. eexists. split; reflexivity.
Qed.

Lemma encode_list_cesd_pair: forall tl hd,
  ~ MValue_is_num (encode_list_cesd (hd :: tl)) ->
  encode_list_cesd (hd::tl) = VP hd (encode_list_cesd tl).
Proof.
intros. simpl in *.
destruct hd; auto.
case_eq (encode_list_cesd tl); auto.
intros. rewrite H0 in H. simpl in H. contradiction H. auto.
Qed.

Lemma relList_le_decr: forall t (tr: irel MValue (CVal t)), 
  ireldecr tr -> ireldecr (relList_le tr).
Proof.
red; intros.
destruct H1 as [vl [tvl [? [? ?]]]].
exists vl. exists tvl.
split.
- generalize tvl H1. clear H1 H2 H3 tvl.
  induction vl; first by [].
  intros. destruct tvl. inversion H1.
  split.
  - apply H with k'. omega. apply H1.
  - apply IHvl. apply H1.
- intuition.
Qed.

(*
   Evaluation
*)

Ltac Lreflect_simpl := Lunfold; reflect_simpl.

Lemma ConsTerm_Ev: forall T (tv1: CVal T) (tv2: CVal (List T)) tv,
  {{ SVal,PVal # tv == subVal [|T|] ConsTerm }} ->
  APP tv (PAIR tv1 tv2) |==> CONS tv1 tv2.
Proof.
move =>  T tv1 tv2 tv Hheq. 
eexists; eapply (UEv_reflects_Ev _ _).
rewrite (Hrw Hheq); Lreflect_simpl.
do 4 (try (simpl; econstructor)).
repeat (Lreflect_simpl); Lsimpl_sub; Lreflect_simpl; by constructor.
Qed.

Lemma SplitTerm_Ev_nil: forall T T1 T2 tv (tv1: CVal T1) (tv2: CVal T2),
  {{ SVal,PVal # tv == subVal [|T|] SplitTerm }} ->
  {{ SVal,PVal # tv1 == NIL [] T }} -> 
  {{ SVal,PVal # tv2 == NONE [] (T *** List T) }} ->
  APP tv tv1 |==> tv2.
Proof.
move => T T1 T2 tv tv1 tv2 Htv Htv1 Htv2.
eexists; eapply (UEv_reflects_Ev _ _).
rewrite -> (Hrw Htv), (Hrw Htv1), (Hrw Htv2).
Lreflect_simpl.
repeat (try (simpl; econstructor)).
- Lreflect_simpl.
  by eapply eq_ind; [constructor | Lreflect_simpl].
- simpl; econstructor.
Qed.

Lemma SplitTerm_Ev_cons: 
  forall T T1 T2 hd tl tv (tv1: CVal T1) (tv2: CVal T2),
  {{ SVal,PVal # tv == subVal [|T|] SplitTerm }} ->
  {{ SVal,PVal # tv1 == CONS (E:=[]) (T:=T) hd (encode_list_src tl) }} ->
  {{ SVal,PVal # tv2 == SOME (PAIR hd (encode_list_src tl)) }} ->
  APP tv tv1 |==> tv2.
Proof.
move => T T1 T2 hd tl tv tv1 tv2 Htv Htv1 Htv2.
rewrite -> (Hrw Htv), (Hrw Htv1), (Hrw Htv2).
eexists; eapply (UEv_reflects_Ev _ _).
Lreflect_simpl.
repeat (try (simpl; econstructor)).
- by Lreflect_simpl; eapply eq_ind; [constructor | Lreflect_simpl].
do 2 Lreflect_simpl. 
match goal with [ |- UEv _ ?tc ?tv] => have: UEv (16 * length tl + 13) tc tv end.
clear Htv Htv1 Htv2; move : hd; induction tl.
- move => hd; simpl; eabs 13.
  repeat (try (simpl; econstructor)).
  by do 2 Lreflect_simpl; constructor. by omega.
- specialize (IHtl a); move => hd; eabs (16 * length (a::tl) + 13).
  do 2 Lreflect_simpl.
  do 2 (simpl; econstructor; first (simpl; econstructor)).
  - match goal with [ |- ?X |- _ => _ ] => eabs X end.
    apply IHtl.
    repeat (simpl; f_equal); try (by do 2 Lreflect_simpl).
  - repeat (simpl; econstructor).
    Lreflect_simpl.
    eapply eq_ind; first by constructor.
    repeat (simpl; f_equal); try (by do 3 Lreflect_simpl).
    by Lsimpl_sub; Lreflect_simpl. by simpl; omega.
move => H; apply H.
Qed.

(************** 
   Correctness 
 **************)

Lemma ListModule_le: #c ListModuleCode <<cc<< ^ListModule.
Proof.
apply cv_implies_cc_le.
unfold ListModule.
eexists; split; first by unfold cstr; apply ObsLeV_refl.
unfold CtoC; red; intros; destruct cesd as [[[c0 e0] s0] d0].
eexists; split; first by eexists; simpl; ksteps.
intros; do 5 red; unfold VtoC; intros; fold irelLeVFun in *.
eapply H4; first last; first (by []); first by apply le_refl.
eexists; split.
- eexists; Hhide_evars; Helimc; Hshow_evars; Lhide; unfold_SubVE; Lshow.
  by apply (@e_TApp (EXIST (#TV 0 *** (#TV 1 *** #TV 0 --> #TV 0) *** 
   (#TV 0 --> UNIT +++ #TV 1 *** #TV 0))) _ t).
- rewrite -> irelLeVFun_fold, !subVal_SubVal; Hsimplc.
  do 3 eexists; split.
  - instantiate (1:= << Sirel # List_eq_ListR t  >> relList_le tr).
    by Hsimpl; Lunfold; Helimc; apply relList_le_decr.
  - split.
    - Lhide; unfold_SubVE; Lshow; rewrite nilMap_idSub SubVal_idSub. 
      by reflexivity.
    - do 2 eexists; split.
      - by Hhide_evars; Hsimplc; Hshow_evars; reflexivity. 
      - split; last (do 2 eexists; split; first (by reflexivity); split).

(* 
   NIL_Src 
*)
simpl; Hsimplc; Lsimpl_sub; Helimc. 
by eexists nil; eexists nil.
   
(* 
   CONS_Src 
*)
eapply eq_ind; first (Hhide_evars; Hrefl (hev0 : SVal PVal); Hshow_evars);
 last (by reflexivity).
Lunfold; Helimc C; intros.
do 4 red; fold irelLeVFun; intros.
destruct v; try inversion H8; rename x into tv1'.
destruct H8 as [tv1 [tv2 [? [? ?]]]]; fold_Sub; subst. 
simpl; ssteps.
have : ((MValue_is_num v1 && MValue_is_num v2) || 
  ~~(MValue_is_num v1 && MValue_is_num v2))
  by (by case : (MValue_is_num v1) (MValue_is_num v2); case).
move /orP; case.

(* v1 is number /\ v2 is number *)
move /andP => Hv1v2.
destruct v1; try by destr_pairs; discriminate.
destruct v2; try by destr_pairs; discriminate.
ssteps.
withcesd c' e' s' d' (destruct (PairNat_obs NPR (c',e',s',d') n n0) as [j' ?]).
eapply stepsfor_kstepreduce_rev; first by simpl in H8; apply H8.
ssteps; eapply H9.
- eexists; split.
  - instantiate (1:= CONS tv1 tv2).
    apply (ConsTerm_Ev tv1 tv2).
    Hsimpl; Lhide; rewrite !isr_sub_cons; Lsimpl_sub.
    replace (isr_sub srel) with (@idsub 0) by mapext_destr.
    rewrite (Hrw (subVal_ss v0 [|t|] (subL idsub))); Hsimpl.
    by rewrite -> consmap_subL, idsub_ss.
  - rewrite irelLeVFun_fold.
    destruct H12 as [? [? [? [vl [tvl [? [? ?]]]]]]]; subst.
    depdestr H12.
    eexists (VN n :: vl); eexists (tv1'::tvl).
    split; first by split; Apply.
    by split; [simpl; rewrite <- H17 | ].
  - omega. - omega.

(* ~(v1 is number /\ v2 is number) *)
move /andP => Hv1v2.
have Hstpv1v2: (stepsfor k3 (c0, e2, VP v1 v2 :: s0, d0)).
  eapply H9.
  - eexists; split.
    - instantiate (1:= CONS tv1 tv2).
      apply (ConsTerm_Ev tv1 tv2).
      Hsimpl; Lhide; rewrite !isr_sub_cons; Lsimpl_sub.
      replace (isr_sub srel) with (@idsub 0) by mapext_destr.
      rewrite (Hrw (subVal_ss v0 [|t|] (subL idsub))); Hsimpl.
      by rewrite -> consmap_subL, idsub_ss.
    - rewrite irelLeVFun_fold.
      destruct H12 as [? [? [? [vl [tvl [? [? ?]]]]]]]; subst.
      depdestr H8.
      eexists (v1 :: vl); eexists (tv1'::tvl).
      split; first by split; Apply.
      split; last by []. 
      simpl; destruct v1; destruct (encode_list_cesd vl); auto.
      by contradiction Hv1v2.
  - omega. - omega.
have : MValue_is_num v1 || ~~ MValue_is_num v1 by destruct v1.
case /orP ; move => Hv1.
- ssteps.
  have : ~ MValue_is_num v2 by (destruct v2; intuition).
  move => Hv2.
  by ssteps.
- move /negP : Hv1 => Hv1.
  by ssteps.

(* 
   SPLIT_Src 
*)
do 4 red; fold irelLeVFun; intros.
rewrite irelLeVFun_fold in H8.
Hhide_concl; revert H8; Hsimpl; Hrefl (tv : SVal PVal).
Lunfold; Helimc C3; Hshow_concl.
intros; repeat red in H8; destruct H8 as [vl [tvl [? [? ?]]]].
apply (eq_Heq SVal PVal) in H13; Hsimpl_hyps.
simpl; ssteps.
have : v = VN 0 \/ ~ v = VN 0
 by (destruct v; first (destruct n); try (right; discriminate); auto).
case; move => Hv.

(* v = VN 0 *)
rewrite Hv in H12; subst.
rewrite (encode_list_cesd_nil _ (sym_eq H12)) in H8.
destruct tvl; last by contradiction.
ssteps; eapply H9.
- eexists; split.
  - eapply SplitTerm_Ev_nil; first 1 last; first (by Apply); first last.
    - rewrite (Hrw (subVal_ss SplitTerm [|t|] (subL (isr_sub srel)))); Hsimpl.
      replace (isr_sub srel) with (@idsub 0) by mapext_destr.
      by rewrite -> consmap_subL, idsub_ss.
    - instantiate (1:= NONE [] _); fold_sub.
      by rewrite -> !isr_sub_cons, List_eq_ListR.
  - fold_sub; repeat red; rewrite irelLeVFun_fold.
    exists (@VUNIT 0 []); by split.
- Apply. - omega.

(* v <> VN 0 *)
ssteps.
rewrite H12 in Hv; destruct vl; first by contradiction Hv.
destruct tvl; first by contradiction H8.
simpl in H8, H13; destruct H8.
have : MValue_is_num v || ~~ MValue_is_num v by destruct (MValue_is_num v).
clear Hv; move /orP; case; [move => Hv | move /negP => Hv].
(* v is num *)
ssteps.
subst; destruct (encode_list_cesd_num _ _ Hv) as [n [n0 [? ?]]]; subst.
clear Hv; simpl; rewrite H15.
ssteps.
withcesd c' e' s' d' 
 (destruct (SplitNat_obs NPR (c',e',s',d') (pair_nat NPR n n0)) as [j' ?]).
simpl in H12; apply (stepsfor_kstepreduce_rev _ H12).
have C : ([],List t) =
 ([]:Env 0, isr_sub (isr_cons (ListR t) 
 (<< Sirel # List_eq_ListR t >> relList_le tr) (isr_cons t tr srel)) (iTVar 1 0))
 by rewrite isr_sub_cons.
ssteps; eapply H9.
- eexists; split.
  - eapply SplitTerm_Ev_cons; first 1 last; first (by Apply); first last.
    - rewrite (Hrw (subVal_ss SplitTerm [|t|] (subL (isr_sub srel)))); Hsimpl.
      replace (isr_sub srel) with (@idsub 0) by mapext_destr.
      by rewrite -> consmap_subL, idsub_ss.
    - instantiate (1:= SOME (PAIR c1 (<<SVal,PVal#C>> encode_list_src tvl))).
      by Hsimpl; Lunfold; Helimc.
  - eexists; split; first by reflexivity.
    eexists; eexists; split; first by reflexivity.
    split; first by rewrite fst_nat_eq; Apply.
    do 3 red; simpl; Hsimpl; Lunfold; Helimc C1.
    rewrite snd_nat_eq.
    eexists; eexists; by split; [apply H14 | split].
- omega. - omega.
(* v is not num *)
ssteps.
subst; replace_hyp Hv (encode_list_cesd_pair _ _ Hv); rewrite Hv.
have C : ([],List t) =
 ([]:Env 0, isr_sub (isr_cons (ListR t) 
 (<< Sirel # List_eq_ListR t >> relList_le tr) (isr_cons t tr srel)) (iTVar 1 0))
 by rewrite isr_sub_cons.
ssteps; eapply H9.
- eexists; split.
  - eapply SplitTerm_Ev_cons; first 1 last; first (by Apply); first last.
    - rewrite (Hrw (subVal_ss SplitTerm [|t|] (subL (isr_sub srel)))); Hsimpl.
      replace (isr_sub srel) with (@idsub 0) by mapext_destr.
      by rewrite -> consmap_subL, idsub_ss.
    - instantiate (1:= SOME (PAIR c1 (<<SVal,PVal#C>> encode_list_src tvl))).
      by Hsimpl; Lunfold; Helimc.
  - eexists; split; first by reflexivity.
    eexists; eexists; split; first by reflexivity.
    split; first by do 3 red; simpl; Apply.
    do 3 red; simpl; Hsimpl; Lunfold; Helimc C1.
    eexists; eexists; by split; [apply H14 | split].
- omega. - omega.
Qed.

Lemma ListModule_ge: #c ListModuleCode >>cc>> ^ListModule.
Proof.
apply cv_implies_cc_ge.
unfold CtoC; exists (fun _:nat => ListModule). 
split; first by unfold cstr; apply ObsLeV_refl. 
move => _; red; intros; destruct cesd as [[[c0 e0] s0] d0]. 
eexists; split; first by eexists; simpl; ksteps.
intros; do 5 red; unfold VtoC; intros; fold erelGeVFun in *.
Hhide_concl; revert H1; Hsimplc.
unfold ListModule; Lhide; simpl; unfold_SubVE; intro H1.
destruct H1 as [k ?]; inversion H1; Hsubst.
revert H2. 
rewrite nilMap_idSub !SubVal_idSub; Hsimplc. 
Lhide. rewrite erelGeVFun_fold. simpl subVal.
rewrite (Hrw (subVal_ss v [|t|] (subL (sr_sub srel)))).
rewrite (Hrw (subVal_ss v0 [|t|] (subL (sr_sub srel)))).
rewrite (Hrw (subVal_ss v1 [|t|] (subL (sr_sub srel)))).
Hsimpl; rewrite -> consmap_subL, idsub_ss; Hsimplc.
Lshow; Hshow_concl.
intros ? H2; apply H2.
exists (ListR t); exists (<< Srel # List_eq_ListR t  >> relList_ge tr); eexists.
split; first by reflexivity.
Hsimplc; intro.
repeat red; rewrite !erelGeVFun_fold.
eexists; eexists; split; first by reflexivity.
split; [ | eexists; eexists; split; first (by reflexivity); split].

(* 
   NIL_Src 
*)
simpl; Hsimplc; Lsimpl_sub; Helimc.
by eexists nil; eexists nil.

(* 
   CONS_Src 
*)

eapply eq_ind; first (Hhide_evars; Hrefl (hev0 : SVal PVal); Hshow_evars);
 last (by reflexivity).
Lunfold; Helimc; intros.
do 4 red; fold erelGeVFun; intros.
move H3 at bottom; destruct v; try inversion H3; rename x into tv1'.
destruct H3 as [tv1 [tv2 [? [? ?]]]]; fold_Sub; subst.
move H6 at bottom; destruct H6 as [tv2' [? [? ?]]].
depdestr H3.
move H7 at bottom; simpl in H7.
move H8 at bottom; simpl in H8; destruct H8 as [vl [tvl [? [? ?]]]].
simpl; tsteps.
have : ((MValue_is_num v1 && MValue_is_num v2) || 
  ~~(MValue_is_num v1 && MValue_is_num v2))
  by (by case : (MValue_is_num v1) (MValue_is_num v2); case).
move /orP; case.

(* v1 is number /\ v2 is number *)
move /andP => Hv1v2.
destruct v1; try by destr_pairs; discriminate.
destruct v2; try by destr_pairs; discriminate.
tsteps.
withcesd c' e' s' d' (destruct (PairNat_obs NPR (c',e',s',d') n n0) as [j' ?]).
eapply terminate_kstepreduce_rev; first by simpl in H11; apply H11.
tsteps; eapply H5.
exists (VN n::vl); exists (tv1'::tvl).
split; first by [].
split; first by simpl; rewrite <- H8.
simpl; rewrite <- H10.
eapply Evv_uniq; first by apply H4.
apply (ConsTerm_Ev tv1' tv2').
Lhide; Hsimpl; rewrite !sr_sub_cons; simpl.
replace (sr_sub srel) with (@idsub 0) by mapext_destr.
by Hunify.

(* ~ (v1 is number /\ v2 is number) *)
move => Hv1v2.
have Hstpv1v2: (terminate (c0, e2, VP v1 v2 :: s0, d0)).
  eapply H5.
  exists (v1::vl); exists (tv1'::tvl).
  split; first by [].
  split.
  - by simpl; rewrite <- H8; destruct v1; destruct v2; auto; discriminate.
  - simpl; rewrite <- H10.
    eapply Evv_uniq; first by apply H4.
    apply (ConsTerm_Ev tv1' tv2').
    Lhide; Hsimpl; rewrite !sr_sub_cons; simpl.
    replace (sr_sub srel) with (@idsub 0) by mapext_destr.
    by Hunify.
move /andP : Hv1v2 => Hv1v2.
have : MValue_is_num v1 || ~~ MValue_is_num v1 by destruct v1.
case /orP ; move => Hv1.
- tsteps.
  have : ~ MValue_is_num v2 by (destruct v2; intuition).  
  move => Hv2.
  by tsteps.
- move /negP : Hv1 => Hv1.
  by tsteps.
 
(* 
   SPLIT_Src 
*)
do 4 red; fold erelGeVFun; intros.
rewrite erelGeVFun_fold in H3.
Hhide_concl; revert H3; Hsimpl; Hrefl (tv : SVal PVal).
Lunfold; Helimc C6; Hshow_concl.
intros; repeat red in H3; destruct H3 as [vl [tvl [? [? ?]]]].
apply (eq_Heq SVal PVal) in H7; Hsimpl_hyps.
simpl; tsteps.
have : v = VN 0 \/ ~ v = VN 0
 by (destruct v; first (destruct n); try (right; discriminate); auto).
case; move => Hv.

(* v = VN 0 *)
rewrite Hv in H6; subst.
rewrite (encode_list_cesd_nil _ (sym_eq H6)) in H3.
destruct tvl; last by contradiction.
tsteps; eapply H5.
eexists; split.
- eapply Evv_uniq; first by apply H4.
  eapply SplitTerm_Ev_nil; first 1 last; first (by Apply); first last.
  - by replace (sr_sub srel) with (@idsub 0) by mapext_destr.
  - by rewrite -> !sr_sub_cons, List_eq_ListR; apply refl_Heq.
- by repeat red.
(* v <> VN 0 *)
tsteps.
rewrite H6 in Hv; destruct vl; first by contradiction Hv.
destruct tvl; first by contradiction H3.
simpl in H3, H7; destruct H3.
have : MValue_is_num v || ~~ MValue_is_num v by destruct (MValue_is_num v).
clear Hv; move /orP; case; [move => Hv | move /negP => Hv].
(* v is num *)
subst; destruct (encode_list_cesd_num _ _ Hv) as [n [n0 [? ?]]]; subst.
clear Hv; simpl; rewrite H9.
tsteps.
withcesd c' e' s' d' 
 (destruct (SplitNat_obs NPR (c',e',s',d') (pair_nat NPR n n0)) as [j' ?]).
simpl in H6; apply (terminate_kstepreduce_rev H6).
have C : ([],List t) =
 ([]:Env 0, sr_sub (sr_cons (ListR t) 
 (<< Srel # List_eq_ListR t >> relList_ge tr) (sr_cons t tr srel)) (iTVar 1 0))
 by rewrite sr_sub_cons.
tsteps; eapply H5.
eexists; split.
- eapply Evv_uniq; first by apply H4.
  eapply SplitTerm_Ev_cons; first 1 last; first (by Apply); first last.
  - by replace (sr_sub srel) with (@idsub 0) by mapext_destr.
  - instantiate (1:= PAIR c1 (<<SVal,PVal#C>> encode_list_src tvl)).
    by Hsimpl; Lunfold; Helimc.
- eexists; eexists; split; first by reflexivity.
  split; first by rewrite fst_nat_eq; Apply.
  do 3 red; simpl; Hsimpl; Lunfold; Helimc.
  rewrite snd_nat_eq.
  eexists; eexists; by split; [Apply | split].
(* v is not num *)
subst; replace_hyp Hv (encode_list_cesd_pair _ _ Hv); rewrite Hv.
have C : ([],List t) =
 ([]:Env 0, sr_sub (sr_cons (ListR t) 
 (<< Srel # List_eq_ListR t >> relList_ge tr) (sr_cons t tr srel)) (iTVar 1 0))
 by rewrite sr_sub_cons.
tsteps; eapply H5.
eexists; split.
- eapply Evv_uniq; first by apply H4.
  eapply SplitTerm_Ev_cons; first 1 last; first (by Apply); first last.
  - by replace (sr_sub srel) with (@idsub 0) by mapext_destr.
  - instantiate (1:= PAIR c1 (<<SVal,PVal#C>> encode_list_src tvl)).
    by Hsimpl; Lunfold; Helimc.
- eexists; eexists; split; first by reflexivity.
  split; first by do 3 red; simpl; Apply.
  do 3 red; simpl; Hsimpl; Lunfold; Helimc.
  eexists; eexists; by split; [Apply | split].
Qed.

Theorem ListModule_eq: #c ListModuleCode ==cc== ^ListModule.
Proof. by split; [apply ListModule_le | apply ListModule_ge]. Qed.

End EgPolList.

