(* 
   Standard operational semantics for \Fv
   Part of the formalization for
   Realizability and Compositional Compiler Correctness 
      for a Polymorphic Language
   N Benton & C-K Hur
   April 2010
*)
Require Import List Program EqNat Omega Bool_nat Heq.
Require Import util sysf.
Require Import ssreflect ssrbool Heq.

Set Implicit Arguments.

(*==========================================================================
  Evaluation relation
  ==========================================================================*)

(** printing =>> %\ensuremath{\Downarrow}% *)
(** printing n1 %\ensuremath{n_1}% *)
(** printing n2 %\ensuremath{n_2}% *)

(* rem printing v1 %\ensuremath{v_1}% *)
(* rem printing v2 %\ensuremath{v_2}% *)
(* rem  printing e1 %\ensuremath{e_1}% *)
(* rem printing e2 %\ensuremath{e_2}% *)

(** printing --> %\ensuremath{\mathrel{\texttt{->}}}% *)
(** printing ** %\ensuremath{\mathrel{\texttt{*}}}% *)
(** printing env %{\ensuremath{\Gamma}}% *)
(** printing env' %{\ensuremath{\Gamma'}}% *)
(** printing ty %{\ensuremath{\tau}}% *)
(** printing ty1 %\ensuremath{\tau_1}% *)
(** printing ty2 %\ensuremath{\tau_2}% *)
(** printing ty' %\ensuremath{\tau'}% *)

(** printing e %{\ensuremath{M}}% *)
(** printing e' %{\ensuremath{M'}}% *)
(** printing e1 %{\ensuremath{M_1}}% *)
(** printing e1' %{\ensuremath{M_1'}}% *)
(** printing e2 %{\ensuremath{M_2}}% *)
(** printing te %{\ensuremath{N}}% *)

(** printing v %{\ensuremath{V}}% *)
(** printing v1 %{\ensuremath{V_1}}% *)
(** printing v2 %{\ensuremath{V_2}}% *)
(** printing tv %{\ensuremath{W}}% *)
(** printing tv' %{\ensuremath{W'}}% *)
(** printing v' %{\ensuremath{V'}}% *)
(** printing {| %{\ensuremath{\{ |}}% *)
(** printing |} %{\ensuremath{| \}}}% *)

(*
   OpSemNat, OpSemBool
*)

Definition OpSemNat (op: OperationNat) : (nat -> nat -> nat) :=
 match op with
 | PLUS => plus
 | MULT => mult
 | MINUS => minus
 end.

Definition OpSemBool (op: OperationBool) : (nat -> nat -> bool) :=
 match op with
 | EQ => fun n n' => beq_nat n n'
 | LE => fun n n' => ble_nat n n'
 | GT => fun n n' => blt_nat n' n
 end.

(*
   Small step operational semantics
*)

Definition TstepC1: forall t (ty:Ty 0),
  (nil : Env 0, subTy [| t |] (subTy shsub ty)) = (nil, ty).
by intros; simpl_sub.
Defined.
Implicit Arguments TstepC1 [t ty].

Inductive Tstep : forall ty, CExp ty -> CExp ty -> Prop :=
(* Val *)
| step_OpN op n1 n2 : 
  Tstep (OPN op (VINT n1) (VINT n2)) (^ VINT (OpSemNat op n1 n2))
| step_OpB op n1 n2 : 
  Tstep (OPB op (VINT n1) (VINT n2)) (^ VBOOL (OpSemBool op n1 n2))
| step_Fst ty1 ty2 (v1 : CVal ty1) (v2 : CVal ty2) :
  Tstep (FST (PAIR v1 v2)) (^ v1)
| step_Snd ty1 ty2 (v1 : CVal ty1) (v2 : CVal ty2) : 
  Tstep (SND (PAIR v1 v2)) (^ v2)

(* Exp *)
| step_CaseL 
  ty ty1 ty2 (v: CVal ty1) (e1: Exp _ [ty1] ty) (e2 : Exp _ [ty2] ty) : 
  Tstep (CASE (INL _ v) e1 e2) (SubExp {|v|} e1)
| step_CaseR 
  ty ty1 ty2 (v: CVal ty2) (e1: Exp _ [ty1] ty) (e2 : Exp _ [ty2] ty) : 
  Tstep (CASE (INR _ v) e1 e2) (SubExp {|v|} e2)
| step_App ty1 ty2 (e: Exp _ [ty1] ty2) (v : CVal ty1) : 
  Tstep (APP (LAM e) v) (SubExp {|v|} e)
| step_RApp ty1 ty2 (e: Exp _ [ty1; ty1 --> ty2] ty2) (v : CVal ty1) : 
  Tstep (APP (REC e) v) (SubExp {|v, REC e|} e)
| step_TApp ty (v: Val 1 nil ty) ty' : 
  Tstep (TAPP (TLAM (env:=nil) v) ty') (^ subVal [|ty'|] v)
| step_Unpk 
  ty1 ty2 ty' (v: CVal (subTy [|ty'|] ty1)) (e: Exp _ [ty1] (subTy shsub ty2)) : 
  Tstep (UNPK (PACK v) e) 
  (<< SExp,PExp # TstepC1 >> SubExp {|v|} (subExp [|ty'|] e))
| step_Let     ty1 ty2 (v : CVal ty1) (e: Exp _ [ty1] ty2) : 
  Tstep (LET (^ v) e) (SubExp {|v|} e)
| step_Cong    ty1 ty2 (e1 e1': CExp ty1) (e2: Exp _ [ty1] ty2) : 
  Tstep e1 e1' -> Tstep (LET e1 e2) (LET e1' e2)
.

Hint Constructors Tstep.

Lemma step_RAppn: 
  forall ty1 ty2 (e: Exp _ [ty1; ty1-->ty2] ty2) (v : CVal ty1) n, 
  Tstep (APP (RECN (S n) e) v) (SubExp {|v, RECN n e|} e).
Proof.
intros; simpl; eapply eq_ind; by [apply step_App | simpl_Sub].
Qed.

Inductive Treduce ty : nat ->  CExp ty -> CExp ty -> Prop :=
  | Tkstep0 tm : Treduce 0 tm tm
  | TkstepS k tm tm1 tm' : 
    Tstep tm tm1 -> Treduce k tm1 tm' -> Treduce (S k) tm tm'
.

Hint Constructors Treduce.

Definition Tstepsfor ty k (tm: CExp ty) := 
  exists k', exists tm', k' >= k /\ Treduce k' tm tm'.

Definition Tdiverge ty (tm: CExp ty) := forall k, Tstepsfor k tm.

Notation "tc |==>!" := (Tdiverge tc) (at level 70, no associativity).

Lemma Tstep_progress: forall ty (tm: CExp ty), 
  (exists tv, tm = ^ tv) \/ (exists tm', Tstep tm tm').
Proof.
cut(forall u env ty tm,
  match u return forall env ty, Exp u env ty -> Prop with 
    | 0 => fun env =>
      match env return forall ty, Exp 0 env ty -> Prop with
        | nil => fun ty tc => 
          (exists tv, tc = ^ tv) \/ (exists tc', Tstep tc tc')
        | t :: E0 => fun _ _ => True
      end
    | _ => fun _ _ _=> True
  end env ty tm).
intros; apply (H _ nil ty tm).

apply Exp_ind2 with (fun _ _ _ _ => True); intros;
  try destruct u; auto; 
  try destruct env; auto;
  try (right; depdestr v; 
    try (inversion v; fail); 
    try (depdestr v0; try (inversion v; fail));
    eexists; constructor).
- by left; eexists.
- right; destruct H as [[? ?] | [? ? ]]; first by rewrite H; eexists.
  by eexists; constructor; Apply.
Qed.

Lemma Tstep_VAL_inv: forall T (tv:CVal T) tc, ~ Tstep (^ tv) tc.
Proof. red; intros; inversion H. Qed.

Lemma Treduce_VAL_inv: forall T k (tv:CVal T) tc,
  Treduce k (^ tv) tc -> k = 0 /\ tc = ^ tv.
Proof.
intros; depdestr H; first by [].
absurd False; auto; eapply Tstep_VAL_inv; apply H.
Qed.

Lemma Tstep_uniq: forall T (tm tm' tm'':CExp T), 
  Tstep tm tm' -> Tstep tm tm'' -> tm' = tm''.
Proof.
intros; induction H;
  try (try (inversion H0; Hsubst); try (inversion H2; Hsubst); auto; fail).
- inversion H0; Hsubst; [ inversion H | f_equal; apply IHTstep; auto ].
Qed.

Lemma Treduce_uniq: forall k T (tm tm' tm'':CExp T), 
  Treduce k tm tm' -> Treduce k tm tm'' -> tm' = tm''.
Proof.
induction k; intros; inversion H; inversion H0; Hsubst; first by [].
rewrite (Tstep_uniq H2 H7) in H3; eapply IHk; Apply.
Qed.

Lemma Treduce_cong: forall k T1 T2 (e1 e1': CExp T1) (e2: Exp _ [T1] T2),
  Treduce k e1 e1' -> Treduce k (LET e1 e2) (LET e1' e2).
Proof.
intros; induction H; first by [].
econstructor; by [constructor; Apply | ].
Qed.

Lemma Treduce_compose: forall k1 k2 T (tm tm1 tm2: CExp T),
  Treduce k1 tm tm1 -> Treduce k2 tm1 tm2 -> Treduce (k1+k2) tm tm2.
Proof.
intros; induction H; first by [].
simpl; econstructor; by [apply H | apply IHTreduce].
Qed.

Lemma Treduce_inv: forall k T (tm tm': CExp T),
  Treduce k tm tm' -> 
  forall k' tm'', Treduce (k+k') tm tm''-> Treduce k' tm' tm''.
Proof.
intros; induction H; first by [].
simpl in H0; depdestr H0.
by apply IHTreduce; rewrite (Tstep_uniq H H0).
Qed.

Lemma Treduce_let_inv: forall k T1 T2 (e1:CExp T1) (e2: Exp 0 [T1] T2) v,
  Treduce k (LET e1 e2) (^ v) ->
  exists k', exists v', k' < k /\ 
    Treduce k' e1 (^ v') /\ Treduce (k-S k') (SubExp {|v'|} e2) (^ v).
Proof.
induction k; intros; first by depdestr H. 
depdestr H, H.
- exists 0; exists v0; split; first by omega. 
  split; by [ apply Tkstep0 | simpl; rewrite <- minus_n_O ].
- destruct (IHk _ _ _ _ _ H0) as [k' [v' [? [? ?]]]].
  exists (S k'); exists v'; split; first by omega. 
  split; by [ eapply TkstepS; by [ apply H | ] | ].
Qed.

Lemma Tstepsfor_zero: forall T (tc:CExp T), Tstepsfor 0 tc.
Proof. intros; do 2 eexists; split; by [ apply le_n | apply Tkstep0 ]. Qed.

Lemma Tstepsfor_step_strict: forall ty k (tm tm':CExp ty),
  Tstep tm tm' -> Tstepsfor (Peano.pred k) tm' -> Tstepsfor k tm.
Proof.
intros; destruct H0 as [k' [tm'' [? ?]]].
exists (S k'); exists tm''; split; first by omega.
eapply TkstepS; by [ apply H | ].
Qed.

Lemma Tstepsfor_step: forall ty k (tm tm':CExp ty),
  Tstep tm tm' -> Tstepsfor k tm' -> Tstepsfor k tm.
Proof.
intros; eapply Tstepsfor_step_strict; first by apply H. 
destruct H0 as [j [tc ?]]; exists j; exists tc; split; by [omega | intuition].
Qed.

(*
   Big step operational semantics, directly
*)
Lemma EvC1: forall t (ty : Ty 0),
  (nil : Env 0, ty) = (nil, subTy [|t|] (subTy shsub ty)).
Proof. intros; simpl_sub. Qed.
Implicit Arguments EvC1 [t ty].

Reserved Notation "e |= i => v" (at level 70, no associativity). 
Inductive Ev: forall ty, nat -> CExp ty -> CVal ty -> Prop :=
| e_Val     ty (v : CVal ty) : ^ v |= 0 => v
| e_OpN     op n1 n2 : 
  OPN op (VINT n1) (VINT n2) |= 1 => VINT (OpSemNat op n1 n2)
| e_OpB     op n1 n2 : 
  OPB op (VINT n1) (VINT n2) |= 1 => VBOOL (OpSemBool op n1 n2)
| e_Fst     ty1 ty2 (v1 : CVal ty1) (v2 : CVal ty2) :
  FST (PAIR v1 v2) |= 1 => v1
| e_Snd     ty1 ty2 (v1 : CVal ty1) (v2 : CVal ty2) :
  SND (PAIR v1 v2) |= 1 => v2
| e_CaseL   k ty ty1 ty2 (v: CVal ty1) e1 e2 (v1: CVal ty) :
  SubExp {|v|} e1 |= k => v1 -> CASE (INL ty2 v) e1 e2 |= S k => v1
| e_CaseR   k ty ty1 ty2 (v: CVal ty2) e1 e2 (v2: CVal ty) :
  SubExp {|v|} e2 |= k => v2 -> CASE (INR ty1 v) e1 e2 |= S k => v2
| e_App     k ty1 ty2 e (v1 : CVal ty1) (v2 : CVal ty2) : 
  SubExp {|v1|} e |= k => v2 -> APP (LAM e) v1 |= S k => v2
| e_RApp    k ty1 ty2 e (v1 : CVal ty1) (v2 : CVal ty2) : 
  SubExp {|v1, REC e|} e |= k => v2 -> APP (REC e) v1 |= S k => v2
| e_TApp    (ty: Ty 1) (v: Val _ nil ty) t :
  TAPP (TLAM (env:=nil) v) t |= 1 => subVal [|t|] v
| e_Unpk    k (ty1: Ty 1) ty2 t (v1: CVal (subTy [|t|] ty1)) v2 
  (e: Exp _ [ty1] (subTy shsub ty2)) : 
  SubExp {|v1|} (subExp [|t|] e) |= k => << SVal,PVal # EvC1 >> v2 ->
  UNPK (PACK v1) e |= S k => v2
| e_Let     k1 k2 ty1 ty2 e1 e2 (v1 : CVal ty1) (v2 : CVal ty2) :
  e1 |= k1 => v1 -> SubExp {|v1|} e2 |= k2 => v2 -> 
  LET e1 e2 |= S (k1 + k2) => v2

where "e |= i => v" := (Ev i e v).

Hint Constructors Ev.

Definition Evv T tc tv := exists k, @Ev T k tc tv.

Notation "e |==> v" := (Evv e v) (at level 70, no associativity).

Definition Terminate T (tc: CExp T) := exists tv, tc |==> tv.

Notation "tc '|==>'" := (Terminate tc) (at level 70, no associativity).

Lemma Ev_to_Treduce: forall i T (tc:CExp T) tv,
  tc |= i => tv -> Treduce i tc (^ tv).
Proof.
intros; induction H; try (by repeat econstructor). 
- repeat econstructor; Hrefl (v2 : SVal PVal); Helimc C. 
  by intros; revert IHEv; Hunify.
- replace (S (k1 + k2)) with (k1 + S k2) by omega.
  eapply Treduce_compose; first by apply Treduce_cong, IHEv1.
  by econstructor; [ apply step_Let | apply IHEv2 ].
Qed.

Lemma Treduce_to_Ev: forall i T (tc:CExp T) tv,
  Treduce i tc (^ tv) -> tc |= i => tv.
Proof.
intro gk; generalize_induction gk; induction gk; intros.
- by inversion H; subst; depdestr H0.
- destruct k; inversion H0; Hsubst; first by [].
  inversion H2; Hsubst;
    try (by apply Treduce_VAL_inv in H3; Hsubst; inversion H1; Hsubst);
    try (by change k with (0+k); econstructor; apply IHgk; auto with arith).

  - econstructor; apply IHgk; auto with arith.
    by revert H3; Hrefl (tv : SVal PVal) at 1; Helimc C0; Hunify.
  - destruct (Treduce_let_inv H3) as [k1 [v' [? [? ?]]]].
    replace k with ((S k1) + (k - (S k1))) by omega.
    econstructor; apply IHgk; try omega; [econstructor; Apply | Apply].
Qed.

Lemma Ev_uniq: forall (ty:Ty 0) (e : CExp ty) k1 v1, e |= k1 => v1 -> 
  forall k2 v2, e |= k2 => v2 -> k1 = k2 /\ v1 = v2.
Proof.
intros; wlog : k1 k2 v1 v2 H H0 / k1 <= k2.
- move => Hassm; destruct (lt_ge_dec k1 k2).
  - apply Hassm; auto with arith.
  - move : (Hassm _ _ _ _ H0 H); intuition.
- intros; apply Ev_to_Treduce in H; apply Ev_to_Treduce in H0.
  replace k2 with (k1 + (k2 - k1)) in * by omega.
  destruct (Treduce_VAL_inv (Treduce_inv H _ H0)).
  depdestr H3; split; by [omega | ].
Qed.

Lemma Evv_uniq: forall (ty:Ty 0) (e : CExp ty) v1, e |==> v1 -> 
  forall v2, e |==> v2 -> v1 = v2.
Proof. by intros; destruct H; destruct H0; apply (proj2 (Ev_uniq H H0)). Qed.

Lemma Ev_step: forall (T:Ty 0) (tm tm':CExp T) tv i,
  Tstep tm tm' -> tm' |= i => tv -> tm |= S i => tv.
Proof.
intros; apply Ev_to_Treduce in H0; apply Treduce_to_Ev. 
by econstructor; Apply.
Qed.

Lemma Evv_step: forall (T:Ty 0) (tm tm':CExp T) tv,
  Tstep tm tm' -> tm' |==> tv -> tm |==> tv.
Proof. by intros; destruct H0; eexists; eapply Ev_step; Apply. Qed.

Lemma Ev_step_rev: forall (T:Ty 0) (tm tm':CExp T) tv i,
  tm |= (S i) => tv -> Tstep tm tm' -> tm' |= i => tv.
Proof.
intros; apply Ev_to_Treduce in H; apply Treduce_to_Ev.
eapply Treduce_inv; by [eapply TkstepS; by [Apply | ] | ].
Qed.

Lemma Evv_step_rev: forall (T:Ty 0) (tm tm':CExp T) tv,
  tm |==> tv -> Tstep tm tm' -> tm' |==> tv.
Proof.
intros; destruct H as [n ?]; destruct n.
- by depdestr H, H0.
- by eexists; eapply Ev_step_rev; Apply.
Qed.

Lemma Ev_VAL_inv: forall T (tv tv':CVal T) i,
  ^ tv |= i => tv' -> tv' = tv.
Proof.
by intros; destruct (Treduce_VAL_inv (Ev_to_Treduce H)); 
 depdestr H1.
Qed.

Lemma Evv_VAL_inv: forall T (tv tv':CVal T),
  ^ tv |==> tv' -> tv' = tv.
Proof. by intros; destruct H; eapply Ev_VAL_inv; Apply. Qed.

Lemma BotF_diverge: forall i T1 T2 (tv : CVal T1),
  Tstepsfor i (APP (BotF nil T1 T2) tv).
Proof.
intros; unfold BotF; induction i.
- apply Tstepsfor_zero.
- by eapply Tstepsfor_step_strict.
Qed.

Lemma Terminate_step: forall T (tc tc': CExp T),
  Tstep tc tc' -> tc' |==> -> tc |==>.
Proof. move => T tc tc' Hstep [tv Hterm]; eexists; eapply Evv_step; Apply. Qed.

Lemma Terminate_step_rev: forall T (tc tc': CExp T),
  tc |==> -> Tstep tc tc' -> tc' |==>.
Proof. 
move => T tc tc' [tv Hterm] Hstep; eexists; eapply Evv_step_rev; Apply. 
Qed.

